/* CPU emulation loop and memory access
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "../libsdl/print.h"
#include "execute.h"
#include "m68k.h"
#include "m68kcpu.h"

//#define DEBUG_CPU

#define IOMASK	0xfffff000
#define IOBASE	0x00dff000

static uint8_t *memory = NULL;
static uint8_t *stop_event = NULL;
static uint32_t memory_size = 0;
static uint16_t *chipreg = NULL;
static uint16_t (*io_read)(uint32_t address) = NULL;
static void (*io_write)(uint32_t address, uint16_t value) = NULL;


static uint16_t read_io_16(unsigned int address)
{
#ifdef DEBUG_CPU
	printf("Chip read from 0x%08x\n", address);
#endif
	return io_read(address);
}

static void write_io_16(unsigned int address, uint16_t value)
{
#ifdef DEBUG_CPU
	printf("Chip write to 0x%08x\n", address);
#endif
	if ((address & IOMASK) == IOBASE)
		chipreg[address & ~IOMASK] = value;
	io_write(address, value);
}

unsigned int m68k_read_memory_8(unsigned int address)
{
	if (address > memory_size - 1) {
		if ((address & 1)) {
			/* read lower (right) byte */
			return read_io_16(address & ~1);
		} else {
			/* read upper (left) byte */
			return read_io_16(address) >> 8;
		}
	}
	return *((uint8_t *)(memory + address));
}

unsigned int m68k_read_memory_16(unsigned int address)
{
	if (address > memory_size - 2) {
		return read_io_16(address);
	}
	return	(memory[address] << 8) |
		 memory[address + 1];
}

unsigned int m68k_read_memory_32(unsigned int address)
{
	if (address > memory_size - 4) {
		int32_t value;
		value = read_io_16(address) << 16;
		value |= read_io_16(address + 2);
		return value;
	}
	return	(memory[address] << 24)|
		(memory[address + 1] << 16) |
		(memory[address + 2] << 8) |
		 memory[address + 3];
}

void m68k_write_memory_8(unsigned int address, unsigned int value)
{
	if (address > memory_size - 1) {
		if ((address & 1)) {
			/* write lower (right) byte */
			write_io_16(address & ~1, 0xff00 | value);
		} else {
			/* write upper (left) byte */
			write_io_16(address, 0x00ff | (value << 8));
		}
		return;
	}
	*((uint8_t *)(memory + address)) = value;
}

void m68k_write_memory_16(unsigned int address, unsigned int value)
{
	if (address > memory_size - 2) {
		write_io_16(address, value);
		return;
	}
	memory[address] = value >> 8;
	memory[address + 1] = value;
}

void m68k_write_memory_32(unsigned int address, unsigned int value)
{
	if (address > memory_size - 4) {
		write_io_16(address, value >> 16);
		write_io_16(address + 2, value);
		return;
	}
	memory[address] = value >> 24;
	memory[address + 1] = value >> 16;
	memory[address + 2] = value >> 8;
	memory[address + 3] = value;
}

void execute_init(int32_t _memory_size, uint8_t *_memory, uint8_t *_stop_event, uint16_t *_chipreg, uint16_t (*_io_read)(uint32_t address), void (*_io_write)(uint32_t address, uint16_t value), const struct cpu_stop stop_at[])
{
	int i;

	memory_size = _memory_size;
	memory = _memory;
	stop_event = _stop_event;
	chipreg = _chipreg;
	io_read = _io_read;
	io_write = _io_write;

	/* flag where the cpu shall stop at */
	memset(stop_event, 0, _memory_size);
	for (i = 0; stop_at[i].event; i++) {
		if (stop_at[i].pc >= memory_size) {
			print_error("stop-event at PC=%x is out of memory=%x\n", stop_at[i].pc, memory_size);
			exit(0);
		}
		stop_event[stop_at[i].pc] = stop_at[i].event;
	}

	/* init CPU */
	m68k_init();
        m68k_set_cpu_type(M68K_CPU_TYPE_68000);
}

void reset_cpu(void)
{
	m68k_pulse_reset();
}

int execute_cpu(int irq, int *event)
{
	int cycle_count = 0;
	int instruction_count = 0;
	uint32_t save_pc;

	if (irq) {
		save_pc = REG_PC;
		m68k_set_irq(irq);
		/* loop until we returned from interrupt */
		do {
			/* execute one opcode */
			cycle_count += m68k_execute(0);
			instruction_count++;
		} while (REG_PC != save_pc);
#ifdef DEBUG_CPU
		printf("interrupt %d took %d opcodes\n", irq, instruction_count);
#endif
	} else {
		do {
			/* execute one opcode */
			cycle_count += m68k_execute(0);
			instruction_count++;

			/* checking if game does not hit any 'stop_at' break point, give error output */
			if (instruction_count >= 10000000) {
				if (instruction_count == 10000000)
					print_error("!!! games seems to got stuck in an endless loop, please fix !!!\n");
				fprintf(stderr, "program counter at: %06x\n", REG_PC);
				if (instruction_count == 10000020)
					break;
			}

			/* stop execution if there is an event */
			if (REG_PC >= memory_size) {
				print_error("CPU execution reaches PC=%x that is out of memory=%x\n", REG_PC, memory_size);
				exit(0);
			}
			if (stop_event[REG_PC]) {
				*event = stop_event[REG_PC];
				return cycle_count;
			}
		} while (42);
	}

	return cycle_count;
}

