/* keyboard emulation
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include "../../include/keycodes.h"
#include "../libsdl/print.h"
#include "keyboard.h"

//#define DEBUG_KEYS

/* stolen from UAE */
static struct keycodes {
	enum keycode key;
	uint8_t code;
} keycodes[] = {
	{ KEYCODE_a, 0x20 },
	{ KEYCODE_b, 0x35 },
	{ KEYCODE_c, 0x33 },
	{ KEYCODE_d, 0x22 },
	{ KEYCODE_e, 0x12 },
	{ KEYCODE_f, 0x23 },
	{ KEYCODE_g, 0x24 },
	{ KEYCODE_h, 0x25 },
	{ KEYCODE_i, 0x17 },
	{ KEYCODE_j, 0x26 },
	{ KEYCODE_k, 0x27 },
	{ KEYCODE_l, 0x28 },
	{ KEYCODE_m, 0x37 },
	{ KEYCODE_n, 0x36 },
	{ KEYCODE_o, 0x18 },
	{ KEYCODE_p, 0x19 },
	{ KEYCODE_q, 0x10 },
	{ KEYCODE_r, 0x13 },
	{ KEYCODE_s, 0x21 },
	{ KEYCODE_t, 0x14 },
	{ KEYCODE_u, 0x16 },
	{ KEYCODE_v, 0x34 },
	{ KEYCODE_w, 0x11 },
	{ KEYCODE_x, 0x32 },
	{ KEYCODE_y, 0x15 },
	{ KEYCODE_z, 0x31 },

	{ KEYCODE_0, 0x0A },
	{ KEYCODE_1, 0x01 },
	{ KEYCODE_2, 0x02 },
	{ KEYCODE_3, 0x03 },
	{ KEYCODE_4, 0x04 },
	{ KEYCODE_5, 0x05 },
	{ KEYCODE_6, 0x06 },
	{ KEYCODE_7, 0x07 },
	{ KEYCODE_8, 0x08 },
	{ KEYCODE_9, 0x09 },

	{ KEYCODE_KP_0, 0x0F },
	{ KEYCODE_KP_1, 0x1D },
	{ KEYCODE_KP_2, 0x1E },
	{ KEYCODE_KP_3, 0x1F },
	{ KEYCODE_KP_4, 0x2D },
	{ KEYCODE_KP_5, 0x2E },
	{ KEYCODE_KP_6, 0x2F },
	{ KEYCODE_KP_7, 0x3D },
	{ KEYCODE_KP_8, 0x3E },
	{ KEYCODE_KP_9, 0x3F },

	{ KEYCODE_KP_DIVIDE, 0x5C },
	{ KEYCODE_KP_MULTIPLY, 0x5D },
	{ KEYCODE_KP_MINUS, 0x4A },
	{ KEYCODE_KP_PLUS, 0x5E },
	{ KEYCODE_KP_PERIOD, 0x3C },
	{ KEYCODE_KP_ENTER, 0x43 },
	{ KEYCODE_KP_LPAREN, 0x5A },
	{ KEYCODE_KP_RPAREN, 0x5B },

	{ KEYCODE_F1, 0x50 },
	{ KEYCODE_F2, 0x51 },
	{ KEYCODE_F3, 0x52 },
	{ KEYCODE_F4, 0x53 },
	{ KEYCODE_F5, 0x54 },
	{ KEYCODE_F6, 0x55 },
	{ KEYCODE_F7, 0x56 },
	{ KEYCODE_F8, 0x57 },
	{ KEYCODE_F9, 0x58 },
	{ KEYCODE_F10, 0x59 },

	{ KEYCODE_UP, 0x4C },
	{ KEYCODE_DOWN, 0x4D },
	{ KEYCODE_LEFT, 0x4F },
	{ KEYCODE_RIGHT, 0x4E },

	{ KEYCODE_SPACE, 0x40 },
	{ KEYCODE_BACKSPACE, 0x41 },
	{ KEYCODE_TAB, 0x42 },
	{ KEYCODE_RETURN, 0x44 },
	{ KEYCODE_ESCAPE, 0x45 },
	{ KEYCODE_DELETE, 0x46 },

	{ KEYCODE_LSHIFT, 0x60 },
	{ KEYCODE_RSHIFT, 0x61 },
	{ KEYCODE_CAPSLOCK, 0x62 },
	{ KEYCODE_LCTRL, 0x63 },
	{ KEYCODE_RCTRL, 0x63 },
	{ KEYCODE_LALT, 0x64 },
	{ KEYCODE_RALT, 0x65 },
	{ KEYCODE_LAMIGA, 0x66 },
	{ KEYCODE_RAMIGA, 0x67 },
	{ KEYCODE_HELP, 0x5F },

	{ KEYCODE_LBRACKET, 0x1A },
	{ KEYCODE_RBRACKET, 0x1B },
	{ KEYCODE_SEMICOLON, 0x29 },
	{ KEYCODE_COMMA, 0x38 },
	{ KEYCODE_PERIOD, 0x39 },
	{ KEYCODE_SLASH, 0x3A },
	{ KEYCODE_BACKSLASH, 0x0D },
	{ KEYCODE_QUOTE, 0x2A },
	{ KEYCODE_NUMBERSIGN, 0x2B },
	{ KEYCODE_LTGT, 0x30 },
	{ KEYCODE_BACKQUOTE, 0x00 },
	{ KEYCODE_MINUS, 0x0B },
	{ KEYCODE_EQUAL, 0x0C },

	{ KEYCODE_UNDEFINED, 0xff },
};

static uint8_t buffer[16];
static int buffer_size = sizeof(buffer), buffer_len = 0;
static uint32_t keydown[4] = {0, 0, 0, 0};

#define ROL(v)		((v << 1) | (v >> 7))

#define CIAASDR		0xbfec00
#define CIAAICR		0xbfed00
#define CIAACRA		0xbfee00

uint16_t emulate_keyboard_read(uint32_t address)
{
	uint8_t value;
	int i;

	switch (address) {
	case CIAAICR:
		if (buffer_len)
			return 0xff08;
		return 0xff00;
	case CIAASDR:
		if (buffer_len == 0)
			return 0xffff;
		value = buffer[0];
#ifdef DEBUG_KEYS
		printf("sending amiga code 0x%02x\n", value);
#endif
		/* rotate and negate */
		value = ~ROL(value);
		/* remove first entry from buffer */
		buffer_len--;
		for (i = 0; i < buffer_len; i++)
			buffer[i] = buffer[i + 1];
		return 0xff00 | value;
	}

	return 0xffff;
}

/* schedule given key to queue,
 * down == 0: up
 * down == 1: down
 */
void set_amiga_key(enum keycode key, int down)
{
	int code = -1, isdown;
	int i;

	for (i = 0; keycodes[i].key != KEYCODE_UNDEFINED; i++) {
		if (keycodes[i].key == key) {
			code = keycodes[i].code;
			break;
		}
	}
	if (code < 0 || code >= 0x80) {
#ifdef DEBUG_KEYS
		printf("Key code '%d' unknown, ignoring!\n", key);
#endif
		return;
	}

	isdown = (keydown[code >> 5] >> (code & 0x1f)) & 1;
	if (down) {
		if (isdown) {
//			printf("key already down\n");
			return;
		}
		keydown[code >> 5] |= (1 << (code & 0x1f));
	} else {
		if (!isdown) {
//			printf("key already up\n");
			return;
		}
		keydown[code >> 5] &= ~(1 << (code & 0x1f));
		code |= 0x80;
	}

	if (buffer_len == buffer_size) {
		print_info("keyboard buffer overflow\n");
		return;
	}

	buffer[buffer_len++] = code;
#ifdef DEBUG_KEYS
	printf("Converting keycode %d to amiga code 0x%02x\n", key, code);
#endif
}


