/* OpenGL rendering
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "../libsdl/print.h"
#include "opengl.h"
#define GLEW_STATIC
#include <GL/glew.h>

#define MAX_OSD 2
static uint8_t *image_rgb = NULL;
static uint8_t *osd_rgba[MAX_OSD] = { NULL, NULL };
static GLuint image_name;
static GLuint osd_name[MAX_OSD];
static int image_width, image_height;
static int osd_width[MAX_OSD], osd_height[MAX_OSD];
static int texture_size;
static int osd_size[MAX_OSD];
static int flip_y;
static GLuint shadow_framebuffer[2] = { 0, 0 };
static int current_view_width = 0, current_view_height = 0;
static GLuint shadow_texture[2];
static int shadow_texture_width = 0, shadow_texture_height = 0;
static int shadow_multisampling;

GLuint current_framebuffer = 0;


/* alloc and init an image texture with size that is greater than the power of two */
int init_opengl_image(int _image_width, int _image_height)
{
	int rc;

	image_width = _image_width;
	image_height = _image_height;

	/* generate texture */
	for (texture_size = 1; texture_size <= image_width || texture_size <= image_height; texture_size *= 2)
		;

	image_rgb = calloc(texture_size * texture_size, 3);
	if (!image_rgb) {
		print_error("Failed to allocate texture\n");
		rc = -ENOMEM;
		goto error;
	}
	glShadeModel(GL_FLAT);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); /* bytes */
	glGenTextures(1, &image_name);
	glBindTexture(GL_TEXTURE_2D, image_name);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, texture_size, texture_size, 0, GL_RGB, GL_UNSIGNED_BYTE, image_rgb);


	return 0;

error:
	exit_opengl();
	return rc;
}

/* alloc and init an osd texture with size that is greater than the power of two */
int init_opengl_osd(int num, int _osd_width, int _osd_height)
{
	int rc;

	if (num < 0 || num >= MAX_OSD) {
		print_error("given OSD number out of range");
		rc = -ENOMEM;
		goto error;
	}

	osd_width[num] = _osd_width;
	osd_height[num] = _osd_height;

	/* generate texture */
	for (osd_size[num] = 1; osd_size[num] <= osd_width[num] || osd_size[num] <= osd_height[num]; osd_size[num] *= 2)
		;
	osd_rgba[num] = calloc(osd_size[num] * osd_size[num], 4);
	if (!osd_rgba[num]) {
		print_error("Failed to allocate texture\n");
		rc = -ENOMEM;
		goto error;
	}
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1); /* bytes */
	glGenTextures(1, &osd_name[num]);
	glBindTexture(GL_TEXTURE_2D, osd_name[num]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, osd_size[num], osd_size[num], 0, GL_RGBA, GL_UNSIGNED_BYTE, osd_rgba[num]);

	return 0;

error:
	exit_opengl();
	return rc;
}

void opengl_clear(int _flip_y)
{
	flip_y = _flip_y;

	/* clear screen */
	glClearColor(0.05, 0.05, 0.05, 1.0);
	glClear(GL_COLOR_BUFFER_BIT);
}

/* only set view width and heigt. (used by ovr to tell frame buffer size) */
void opengl_set_size(int view_width, int view_height)
{
	current_view_width = view_width;
	current_view_height = view_height;
}

/* set viewport for improved rendering */
void opengl_viewport(int view_width, int view_height, int split, int benson_at_line, double fov, double benson_size)
{
	int view_x = 0, view_y = 0;
	double factor_height = 1.0;
	double factor_width = 1.0;

	/* center view, or put it in the top half of the window */
	if (split == 1) {
		view_y = view_height / 2;
		view_height = view_height / 2;
	} else if (split == 2) {
		view_height = view_height / 2;
	}

	/* avoid division by zero, if one dimension is too small */
	if (view_width < 1 || view_height < 1)
		return;

	/* expand width or height, if apect of image_width:image_height does not match the view_width:view_height */
	if (view_height * image_width > view_width * image_height)
		factor_height = (double)(view_height * image_width) / (double)(view_width * image_height);
	else if (view_height * image_width < view_width * image_height)
		factor_width = (double)(view_width * image_height) / (double)(view_height * image_width);

	current_view_width = view_width;
	current_view_height = view_height;

	/* avoid views that are too small */
	if (view_width < 1 || view_height < 1)
		return;

	/* projection matrix */
	glViewport((GLsizei)view_x, (GLsizei)view_y, (GLsizei)view_width, (GLsizei)view_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	/* calculate field-of-view */
	double slope = tan(fov / 360 * M_PI);
	/* make frustum to center the view in the game view above benson */
	double left = -slope;
	double right = slope;
	double benson_start_at_position = ((double)image_height - (double)benson_at_line) * benson_size;
	double top = slope * ((double)image_height - benson_start_at_position) / (double)image_width;
	double bottom = -slope * ((double)image_height * 2.0 - ((double)image_height - benson_start_at_position)) / (double)image_width;
	glFrustum(left * factor_width, right * factor_width, bottom * factor_height, top * factor_height, 1.0, 5000000000.0);

	glMatrixMode(GL_MODELVIEW);
}

/* render image or only benson */
void opengl_blit_image(uint8_t *rgb, int filter, int benson_at_line, int render_benson_only, double fov, double monitor_distance, double benson_size, int benson_case)
{
	double texture_left = 0;
	double texture_right = ((double)image_width) / (double)texture_size;
	double texture_top = ((double)benson_at_line) / (double)texture_size;
	double texture_bottom = ((double)image_height) / (double)texture_size;
	double image_start_at_position = ((double)image_height - (double)benson_at_line) * benson_size;
	double image_top = -((double)image_height - image_start_at_position) / (double)image_width;
	double image_bottom = -((double)image_height * 2.0 - ((double)image_height - image_start_at_position)) / (double)image_width;

	/* calculate field-of-view */
	double slope = tan(fov / 360 * M_PI) * monitor_distance;

	/* render image */

	if (benson_case) {
		glEnable(GL_CULL_FACE);
		glFrontFace(GL_CW);
		glCullFace(GL_BACK);
	}

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, image_name);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);  /* no modulation with color */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, (filter) ? GL_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (filter) ? GL_LINEAR : GL_NEAREST);
	if (render_benson_only) {
		/* render benson only, and copy top line of benson to one line above, so the texture filter will not take false data above benson */
		rgb += image_width * (benson_at_line - 1) * 3;
		memcpy(rgb, rgb + image_width * 3, image_width * 3);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, benson_at_line - 1, image_width, image_height - benson_at_line - 1, GL_RGB, GL_UNSIGNED_BYTE, rgb);
	} else {
		image_top = -image_top;
		texture_top = 0.0;
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, image_width, image_height, GL_RGB, GL_UNSIGNED_BYTE, rgb);
	}
	glBegin(GL_QUADS);
	glTexCoord2f(texture_left, texture_bottom);
	glVertex3f(-benson_size * slope, image_bottom * slope, -monitor_distance);
	glTexCoord2f(texture_right, texture_bottom);
	glVertex3f(benson_size * slope, image_bottom * slope, -monitor_distance);
	glTexCoord2f(texture_right, texture_top);
	glVertex3f(benson_size * slope, image_top * slope, -monitor_distance);
	glTexCoord2f(texture_left, texture_top);
	glVertex3f(-benson_size * slope, image_top * slope, -monitor_distance);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	/* render benson case */
	if (benson_case) {
		double left, right, bottom, top, near, far;
		left = -benson_size * slope;
		right = benson_size * slope;
		bottom = image_bottom * slope;
		top = image_top * slope;
		near = -monitor_distance;
		far = near - benson_size * slope / 3.0;
		glBegin(GL_QUADS);
		/* back */
		glColor4f(0.3, 0.3, 0.3, 1.0);
		glVertex3f(left, top, far);
		glVertex3f(right, top, far);
		glVertex3f(right, bottom, far);
		glVertex3f(left, bottom, far);
		/* top */
		glColor4f(0.2, 0.2, 0.2, 1.0);
		glVertex3f(left, top, near);
		glVertex3f(right, top, near);
		glVertex3f(right, top, far);
		glVertex3f(left, top, far);
		/* bottom */
		glColor4f(0.05, 0.05, 0.05, 1.0);
		glVertex3f(left, bottom, far);
		glVertex3f(right, bottom, far);
		glVertex3f(right, bottom, near);
		glVertex3f(left, bottom, near);
		/* left */
		glColor4f(0.1, 0.1, 0.1, 1.0);
		glVertex3f(left, bottom, far);
		glVertex3f(left, bottom, near);
		glVertex3f(left, top, near);
		glVertex3f(left, top, far);
		/* right */
		glVertex3f(right, top, far);
		glVertex3f(right, top, near);
		glVertex3f(right, bottom, near);
		glVertex3f(right, bottom, far);
		glEnd();
	}

	if (benson_case) {
		glDisable(GL_CULL_FACE);
	}
}

/* render osd texture */
void opengl_blit_osd(int num, uint8_t *rgba, int filter, int benson_at_line, double fov, double monitor_distance, double benson_size, double scale_x, double scale_y, double offset_x, double offset_y)
{
	double texture_left = 0.0;
	double texture_right = (double)osd_width[num] / (double)osd_size[num];
	double texture_top = 0.0;
	double texture_bottom = (double)osd_height[num] / (double)osd_size[num];
	double benson_start_at_position;
	double osd_left = 0.0;
	double osd_right = 1.0;
	double osd_top = 0.0;
	double osd_bottom = 1.0;
	double slope, range, center;
	if (fov) {
		osd_left = -1.0;
		osd_right = 1.0;
		benson_start_at_position = ((double)image_height - (double)benson_at_line) * benson_size;
		osd_top = ((double)image_height - benson_start_at_position) / (double)image_width;
		osd_bottom = -((double)image_height * 2.0 - ((double)image_height - benson_start_at_position)) / (double)image_width;
		/* calculate field-of-view */
		slope = tan(fov / 360 * M_PI) * monitor_distance;
	}
	range = (osd_right - osd_left) / 2.0;
	center = (osd_right + osd_left) / 2.0;
	osd_left = (osd_left - center) * scale_x + center + range * offset_x;
	osd_right = (osd_right - center) * scale_x + center + range * offset_x;
	range = (osd_bottom - osd_top) / 2.0;
	center = (osd_bottom + osd_top) / 2.0;
	osd_top = (osd_top - center) * scale_y + center + range * offset_y;
	osd_bottom = (osd_bottom - center) * scale_y + center + range * offset_y;

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, osd_name[num]);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);  /* no modulation with color */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, (filter) ? GL_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (filter) ? GL_LINEAR : GL_NEAREST);
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, osd_width[num], osd_height[num], GL_RGBA, GL_UNSIGNED_BYTE, rgba);
	glBegin(GL_QUADS);
	if (fov) {
		/* perspective viewport */
		glTexCoord2f(texture_left, texture_bottom);
		glVertex3f(osd_left * slope, osd_bottom * slope, -monitor_distance);
		glTexCoord2f(texture_right, texture_bottom);
		glVertex3f(osd_right * slope, osd_bottom * slope, -monitor_distance);
		glTexCoord2f(texture_right, texture_top);
		glVertex3f(osd_right * slope, osd_top * slope, -monitor_distance);
		glTexCoord2f(texture_left, texture_top);
		glVertex3f(osd_left * slope, osd_top * slope, -monitor_distance);
	} else {
		/* orthogonal viewport */
		glTexCoord2f(texture_left, texture_top);
		glVertex3f(osd_left, osd_top, 0.0);
		glTexCoord2f(texture_right, texture_top);
		glVertex3f(osd_right, osd_top, 0.0);
		glTexCoord2f(texture_right, texture_bottom);
		glVertex3f(osd_right, osd_bottom, 0.0);
		glTexCoord2f(texture_left, texture_bottom);
		glVertex3f(osd_left, osd_bottom, 0.0);
	}
	glEnd();
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
}

/* set color and opacity */
void opengl_render_color(double r, double g, double b, double a)
{
	if (a < 1.0) {
		glEnable(GL_BLEND);
	       	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glColor4d(r, g, b, a);
	} else {
		glDisable(GL_BLEND);
		glColor3d(r, g, b);
	}
}

/* set color and alpha */
void opengl_render_color_alpha(double r, double g, double b, double a)
{
	glDisable(GL_BLEND);
	glColor4d(r, g, b, a);
}

/* render polygon */
void opengl_render_polygon(double *x, double *y, double *z, int count, int cull_face)
{
	int i;

	if (cull_face) {
	        glEnable(GL_CULL_FACE);
	        glFrontFace((flip_y) ? GL_CCW : GL_CW);
	        glCullFace(GL_BACK);
	}
	glBegin(GL_POLYGON);
	for (i = 0; i < count; i++)
		glVertex3d(x[i], y[i], -z[i]);
	glEnd();
	if (cull_face)
		glDisable(GL_CULL_FACE);
}

/* render polygon, but make sure any size of it is visible */
void opengl_render_polygon_and_line(double *x, double *y, double *z, int count)
{
	int i;

	glBegin(GL_POLYGON);
	for (i = 0; i < count; i++)
		glVertex3d(x[i], y[i], -z[i]);
	glEnd();
	glBegin(GL_LINE_LOOP);
	for (i = 0; i < count; i++)
		glVertex3d(x[i], y[i], -z[i]);
	glEnd();
}

/* render line */
void opengl_render_line(double x1, double y1, double z1, double x2, double y2, double z2, double size)
{
	if (size == 0.0) {
		glBegin(GL_LINES);
		glVertex3f(x1, y1, -z1);
		glVertex3f(x2, y2, -z2);
		glEnd();
		return;
	}
}

/* render point */
void opengl_render_point(double x, double y, double z, double size)
{
	if (size == 0.0) {
		glBegin(GL_POINTS);
		glVertex3f(x, y, -z);
		glEnd();
		return;
	}
}

int init_shadow_buffer(int multisampling)
{
	shadow_multisampling = multisampling;

	glGenFramebuffers(2, shadow_framebuffer);

	glBindFramebuffer(GL_FRAMEBUFFER, current_framebuffer);

	return 0;
}

/* change to shadow buffer and clear it */
void begin_shadow_render(void)
{
	/* free texture, if size changes */
	if (shadow_texture[0] && (current_view_width != shadow_texture_width || current_view_height != shadow_texture_height)) {
		glDeleteTextures(2, shadow_texture);
		shadow_texture[0] = 0;
	}
	/* generate texture for shadow rendering (two, if MSAA is used) */
	if (!shadow_texture[0]) {
		glGenTextures(2, shadow_texture);
		shadow_texture_width = current_view_width;
		shadow_texture_height = current_view_height;
		if (shadow_multisampling > 1) {
			glBindTexture(GL_TEXTURE_2D_MULTISAMPLE, shadow_texture[0]);
			glTexImage2DMultisample(GL_TEXTURE_2D_MULTISAMPLE, shadow_multisampling, GL_RGBA, shadow_texture_width, shadow_texture_height, 0);
			glBindTexture(GL_TEXTURE_2D, shadow_texture[1]);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, shadow_texture_width, shadow_texture_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		} else {
			glBindTexture(GL_TEXTURE_2D, shadow_texture[0]);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, shadow_texture_width, shadow_texture_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
		}
	}

	/* switch to shadow_framebuffer now and attach the first texture */
	glBindFramebuffer(GL_FRAMEBUFFER, shadow_framebuffer[0]);
	if (shadow_multisampling > 1)
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D_MULTISAMPLE, shadow_texture[0], 0);
	else
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, shadow_texture[0], 0);
	GLenum DrawBuffers[1] = {GL_COLOR_ATTACHMENT0};
	glDrawBuffers(1, DrawBuffers);
	if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
		fprintf(stderr, "failed to create shadow texture\n");
	}

	/* clear to black + transparent */
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT);
}

/* apply shadow from shadow_framebuffer to current_framebuffer and switch to it */
void end_shadow_render(void)
{
	if (shadow_multisampling > 1) {
		/* attach second texture to second frame buffer */
		glBindFramebuffer(GL_FRAMEBUFFER, shadow_framebuffer[1]);
		glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, shadow_texture[1], 0);
		glBindTexture(GL_TEXTURE_2D_MULTISAMPLE, shadow_texture[1]);
		/* blit from first texture to second texture (MSAA to normal texture) */
		glBindFramebuffer(GL_READ_FRAMEBUFFER, shadow_framebuffer[0]);
		glBindFramebuffer(GL_DRAW_FRAMEBUFFER, shadow_framebuffer[1]);
		glBlitFramebufferEXT( 0, 0, shadow_texture_width, shadow_texture_height, 0, 0, shadow_texture_width, shadow_texture_height, GL_COLOR_BUFFER_BIT, GL_NEAREST);
	}

	/* switch to color buffer */
	glBindFramebuffer(GL_FRAMEBUFFER, current_framebuffer);

	/* apply shadow by drawing texture to screen */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4d(1, 1, 1, 1);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(-1, 1, -1, 1, -1, 1);

	glEnable(GL_TEXTURE_2D);
	if (shadow_multisampling > 1)
		glBindTexture(GL_TEXTURE_2D, shadow_texture[1]);
	else
		glBindTexture(GL_TEXTURE_2D, shadow_texture[0]);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 1);
	glVertex3f(-1, 1, 0);
	glTexCoord2f(1, 1);
	glVertex3f(1, 1, 0);
	glTexCoord2f(1, 0);
	glVertex3f(1, -1, 0);
	glTexCoord2f(0, 0);
	glVertex3f(-1, -1, 0);
	glEnd();
	glDisable(GL_TEXTURE_2D);

	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glDisable(GL_BLEND);
}

/* free image texture */
void exit_opengl(void)
{
	int i;

	if (shadow_texture[0]) {
		glDeleteTextures(2, shadow_texture);
		shadow_texture[0] = 0;
	}
	if (shadow_framebuffer[0]) {
		glDeleteFramebuffers(2, shadow_framebuffer);
		shadow_framebuffer[0] = 0;
	}
	if (image_rgb) {
		free(image_rgb);
		image_rgb = NULL;
	}
	for (i = 0; i < MAX_OSD; i++) {
		if (osd_rgba[i]) {
			free(osd_rgba[i]);
			osd_rgba[i] = NULL;
		}
	}
}

