/* render routines that replaces the game rendering by OpenGL rendering
 *
 * (C) 2018 by Andreas Eversberg <jolly@eversberg.eu>
 * All Rights Reserved
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* How it works:
 *
 * 1. Each game's rendering is capured:
 *    a: call to render_capture_start()
 *    b: several calls to render_capture_event() to capture all data
 *    c: call to render_capture_stop()
 * 2. The captured motion is then interpolated to get smooth motion:
 *    - new_motion for motion of current/recent capture
 *    - old_motion for motion of previous capture
 *    - interpolation for interpolated result
 * 3. The complete scene is rendered:
 *    - render_all_items() calls the render_item() for each item
 *    - All items before shadows on the ground are rendered first
 *    - Then all items with shadows are rendered
 *    - Then all items after shadows are rendered
 *    - The recent capture (NEW) is rendered
 *    - Interpolation result is taken into account
 *
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../libsdl/print.h"
#include "../libcpu/m68k.h"
#include "../libcpu/m68kcpu.h"
#include "../libcpu/execute.h"
#include "../libopengl/opengl.h"
#include "mercenary.h"

#define GL3_PROTOTYPES 1
#include <GL/glew.h>

//#define DEBUG_COLOR
//#define DEBUG_VERTEX
//#define DEBUG_ITEMS
//#define DEBUG_LIGHT

#define ELEVATION_SKY_NIGHT	-0.05	/* elevation of darkest sky */
#define ELEVATION_SKY_DAY	0.3	/* elevation of brightest sky */
#define ELEVATION_STARS_NIGHT	-0.1	/* elevation of brightest stars */
#define ELEVATION_STARS_DAY	0.1	/* elevation of no stars */
#define ELEVATION_SHADOW_MIN	0.03	/* elevation of shadow begin */
#define ELEVATION_SHADOW_MAX	0.3	/* elevation of darkest shadow */
#define SHADOW_LEVEL		0.9	/* opacity of darkest shadow */

#define MAX_POLYGON		16	/* number of polygon corners (vertices) */
#define MAX_VERTEX		0x100	/* this is the value range, these are 64 vertices */
#define MAX_INTERIOR_VERTEX	0x400	/* do we need that much? */
#define MAX_INTERSTARS		80	/* always 80 stars */
#define MAX_MOVING_OBJECTS	16	/* maximum number of moving objects (used for interpolation) */
#define MAX_EXPLOSION		256	/* how many explosion particles can be stored in one object */
#define PLANET_VERTICES		128
#define PLANET_STRETCH		1.20	/* stretch horizontally (this compensates NTSC/PAL line ratio of 576/480)*/
#define EXPLOSION_VERTICES	16
#define EXPLOSION_STRETCH	1.20	/* as above */
#define SIGHT_DIST		78.74	/* distanc of sights in inch */
#define FIX_OBJECT_SCALE	16	/* intercity... */
#define FIX_OBJECT_SCALE_TAG	2	/* line in faces of persons */

enum render_shadow {
	RENDER_BEFORE_SHADOW,
	RENDER_AT_SHADOW,
	RENDER_AFTER_SHADOW,
};

/*
 *  render item definition and structures
 */

/* render items */
enum render_item_type {
	RENDER_ITEM_OBJECT_INFO,
	RENDER_ITEM_VERTICES_0,
	RENDER_ITEM_VERTICES_1,
	RENDER_ITEM_VERTICES_2,
	RENDER_ITEM_VERTICES_INTERIOR,
	RENDER_ITEM_SKY,
	RENDER_ITEM_GROUND,
	RENDER_ITEM_OBJECT_POLYGON,
	RENDER_ITEM_OBJECT_LINE,
	RENDER_ITEM_BEACON_POINT,
	RENDER_ITEM_BUILDING_EXTERIOR_POLYGON,
	RENDER_ITEM_BUILDING_EXTERIOR_LINE,
	RENDER_ITEM_BUILDING_INTERIOR_1TO4,
	RENDER_ITEM_BUILDING_INTERIOR_5TO6,
	RENDER_ITEM_BUILDING_INTERIOR_WALL,
	RENDER_ITEM_COMET_POLYGON,
	RENDER_ITEM_ROAD_LINE,
	RENDER_ITEM_ROAD_POLYGON,
	RENDER_ITEM_TAG_LINE_OBJECT,
	RENDER_ITEM_TAG_LINE_OTHER,
	RENDER_ITEM_TAG_POLYGON_OBJECT,
	RENDER_ITEM_TAG_POLYGON_OTHER,
	RENDER_ITEM_PLANET,
	RENDER_ITEM_STARS,
	RENDER_ITEM_INTERSTELLAR_STARS,
	RENDER_ITEM_INTERSTELLAR_SUN,
	RENDER_ITEM_ISLAND_POLYGON,
	RENDER_ITEM_SIGHTS,
	RENDER_ITEM_EXPLOSION,
	RENDER_ITEM_SHADOW_BUILDING_POLYGON,
	RENDER_ITEM_SHADOW_BUILDING_LINE,
	RENDER_ITEM_SHADOW_OBJECT_POLYGON,
	RENDER_ITEM_SHADOW_OBJECT_LINE,
};

struct render_item_info {
	int moving;
	int id;
	int32_t east, height, north;
};

struct render_item_vertices {
	double x[MAX_VERTEX >> 2], y[MAX_VERTEX >> 2], z[MAX_VERTEX >> 2];
};

struct render_item_vertices_interior {
	uint8_t set[MAX_INTERIOR_VERTEX >> 2];
	double x[MAX_INTERIOR_VERTEX >> 2], y[4], z[MAX_INTERIOR_VERTEX >> 2];
};

struct render_item_sky {
	double red, green, blue;
};

struct render_item_ground {
	double red, green, blue;
};

struct render_item_polygon {
	double red, green, blue;
	int vertices;
	int vertex[MAX_POLYGON];
};

struct render_item_interior14 {
	double red, green, blue;
	int level;
	int vertex[4];
};

struct render_item_interior56 {
	double red, green, blue;
	int level12;
	int level34;
	int vertex14;
	int vertex23;
};

struct render_item_line {
	double red, green, blue;
	int vertex[2];
};

struct render_item_point {
	double red, green, blue;
	int vertex;
};

struct render_item_planet {
	double front_red, front_green, front_blue;
	double back_red, back_green, back_blue;
	int vertex;
	double size;
};

struct render_item_stars {
	int16_t v_offset;
	int tilt;
	int32_t tilt_value;
	int above_zenith;
};

struct render_item_interstars {
	uint8_t color[MAX_INTERSTARS];
	int16_t x[MAX_INTERSTARS], y[MAX_INTERSTARS];
	int	count;
};

struct render_item_explosion {
	double red, green, blue;
	int32_t x[MAX_EXPLOSION], y[MAX_EXPLOSION], z[MAX_EXPLOSION];
	double size[MAX_EXPLOSION];
	int count;
};

struct render_item_shadow_polygon {
	int32_t x_pos, y_pos, z_pos;
	int vertices;
	int32_t x[MAX_POLYGON], y[MAX_POLYGON], z[MAX_POLYGON];
};

struct render_item_shadow_line {
	int32_t x_pos, y_pos, z_pos;
	int32_t x[2], y[2], z[2];
};

typedef struct render_item {
	struct render_item *next;
	enum render_item_type type;
	union {
		struct render_item_info			info;
		struct render_item_vertices		vertices;
		struct render_item_vertices_interior	vertices_interior;
		struct render_item_sky			sky;
		struct render_item_ground		ground;
		struct render_item_polygon		polygon;
		struct render_item_line			line;
		struct render_item_point		point;
		struct render_item_interior14		interior14;
		struct render_item_interior56		interior56;
		struct render_item_planet		planet;
		struct render_item_stars		stars;
		struct render_item_interstars		interstars;
		struct render_item_explosion		explosion;
		struct render_item_shadow_polygon	shadow_polygon;
		struct render_item_shadow_line		shadow_line;
	} u;
} render_item_t;

/* information about motion in each game rendering */
typedef struct motion {
	int32_t position_east, position_height, position_north;
	double orientation_roll, orientation_pitch, orientation_yaw;
	uint16_t orientation_raw_yaw;
	int16_t orientation_raw_pitch;
	int planet_rotation;
	double planet_inclination, planet_azimuth;
} motion_t;

/* information about interpolation between two game renedrings */
typedef struct interpolation {
	double offset_east, offset_height, offset_north;
	double orientation_roll, orientation_pitch, orientation_yaw;
	double orientation_raw_yaw, orientation_raw_pitch;
	double planet_inclination, planet_azimuth;
	int object_id[MAX_MOVING_OBJECTS];
	double object_offset_east[MAX_MOVING_OBJECTS], object_offset_height[MAX_MOVING_OBJECTS], object_offset_north[MAX_MOVING_OBJECTS];
	int object_count;
	render_item_t *interior;
	render_item_t *planets;
} interpolation_t;

#define GET_ORIENTATION \
	double roll = interpolation.orientation_roll; \
	double pitch = interpolation.orientation_pitch; \
	double yaw = interpolation.orientation_yaw

#define GET_ORIENTATION_FIX \
	roll = motion_new.orientation_roll; \
	pitch = motion_new.orientation_pitch; \
	yaw = motion_new.orientation_yaw

/* rendering options */
static int extend_roads; /* extend roads in its original width, rather than just using a single point */
static int improve_stars; /* stars are rendered spherical */
static int fix_sky_rotation; /* sky will rotate correctly by rotating planets/comet by 180 degrees */
static int draw_shadows; /* render shadows */
static double planet_aspect, explosion_aspect; /* aspect ratio */
static double fov;
static double debug_opacity;
static double frustum_slope_64, frustum_slope_fov;

/* states while collecting render items */
static motion_t motion_old, motion_new;
static int32_t old_height_offset = 0, new_height_offset = 0;
static interpolation_t interpolation;
static int ground_index, last_ground_index = -1;
static int interior_level12 = 0;
static int interior_level34 = 0;
static int tag_is_object;
/* current render item list */
static render_item_t *render_list_new = NULL, **render_list_end = &render_list_new;
/* previous render item list */
static render_item_t *render_list_old = NULL;
/* current item to be processed */
static render_item_t *render_item;
/* current object info */
static render_item_t *render_item_object_info;
/* current vertices */
static render_item_t *render_item_vertices_0, *render_item_vertices_1, *render_item_vertices_2;
static render_item_t *render_item_vertices_planets, *render_item_vertices_interior;

/* states while rendering */
int sun_valid = 0;
double stars_bright; /* the brightness of stars */
double shadow_level; /* how 'dark' the shadow is (that is darker at day) */
double sun_pos_x, sun_pos_y, sun_pos_z;
double sky_color_red, sky_color_green, sky_color_blue;

/*
 * capturing
 */

static void render_item_add(enum render_item_type type)
{
	render_item = calloc(1, sizeof(render_item_t));
	if (!render_item) {
		print_error("No memory, must abort!\n");
		abort();
	}
	render_item->type = type;
	*render_list_end = render_item;
	render_list_end = &render_item->next;
}

static void flush_old_items(void)
{
	/* flush old render list */
	while (render_list_old) {
		render_item = render_list_old;
		render_list_old = render_list_old->next;
		free(render_item);
	}
}

/* rendering starts, initialize variables */
void render_capture_start(double _fov, int _extend_roads, int _smooth_planets, int _improve_stars, int _fix_sky_rotation, int _round_planets, int _shadows, int debug)
{
#if defined(DEBUG_COLOR) || defined(DEBUG_VERTEX)
	printf("start rendering a new frame...\n");
#endif

	flush_old_items();
	render_item = NULL;
	/* move new render list to old render list */
	render_list_old = render_list_new;
	/* setup new render list */
	render_list_new = NULL;
	render_list_end = &render_list_new;

	/* move new motion to old motion */
	memcpy(&motion_old, &motion_new, sizeof(motion_old));

	/* set rendering options */
	fov = _fov;
	extend_roads = _extend_roads;
	fix_sky_rotation = _fix_sky_rotation;
	improve_stars = _improve_stars;
	planet_aspect = (_round_planets) ? 1.0 : PLANET_STRETCH;
	explosion_aspect = (_round_planets) ? 1.0 : EXPLOSION_STRETCH;
	draw_shadows = _shadows;
	/* set some transpareny, if debugging is enabled */
	debug_opacity = (debug) ? 0.5 : 1.0;

	/* calculate slope of 64 degree frustum and current FOV's frustum */	
	frustum_slope_64 = tan(64.0 / 2.0 / 180.0 * M_PI);
	frustum_slope_fov = tan(fov / 2.0 / 180.0 * M_PI);

	/* init motion */
	mercenary_get_location(&motion_new.position_east, &motion_new.position_height, &motion_new.position_north);
	/* in case of player on the ground, there is no roll, so it will be reset to 0 when screen is cleared */
	mercenary_get_orientation(&motion_new.orientation_roll, &motion_new.orientation_pitch, &motion_new.orientation_yaw);
	mercenary_get_orientation_raw(&motion_new.orientation_raw_pitch, &motion_new.orientation_raw_yaw);
	motion_new.planet_rotation = motion_old.planet_rotation;
	mercenary_get_orientation_planet(&motion_new.planet_inclination, &motion_new.planet_azimuth, _smooth_planets);

	render_item_object_info = NULL;
	render_item_vertices_0 = render_item_vertices_1 = render_item_vertices_2 = NULL;
	render_item_vertices_planets = NULL;
	render_item_vertices_interior = NULL;

	/* detect elevator movement */
	old_height_offset = new_height_offset;
	if (render_list_old)
		new_height_offset = motion_new.position_height - motion_old.position_height;

	/* detect switching between space (-1) and over ground (>=0) */
	last_ground_index = ground_index;

	/* init tag type */
	tag_is_object = 0;
}

void render_capture_stop(void)
{
	/* no new list (due to STOP_AT_WAIT_INPUT), use old list, if any */
	if (!render_list_new) {
		render_list_new = render_list_old;
		render_list_end = &render_list_new;
		render_list_old = NULL;
	}

}

static void gamecolor2gl_index(double *red, double *green, double *blue, uint16_t index)
{
	uint32_t palette;
	uint16_t color;

	/* use given index from current palette, so we take the palette from M3:DS_0+0x1FA8 */
	index <<= 1;
	palette = mercenary_palette_render();
	color = m68k_read_memory_16(palette + index);
#ifdef DEBUG_COLOR
	printf("using color index (%d) from current palette, color is now 0x%04x\n", index, color);
#endif
	if (color >= 0x8000) {
#ifdef DEBUG_COLOR
		fprintf(stderr, "Use of color index from current palette, but index is not defined as being set!\n");
#endif
	}
	*red = (double)((color >> 8) & 0xf) / 15.0;
	*green = (double)((color >> 4) & 0xf) / 15.0;
	*blue = (double)(color & 0xf) / 15.0;
}

static void gamecolor2gl(double *red, double *green, double *blue, uint16_t color)
{
	uint16_t index;
	uint32_t palette;
	int nesting = 0;

#ifdef DEBUG_COLOR
	printf("color is given as 0x%04x\n", color);
#endif
again:
	/* color conversion: see for example M3: 0x4f830 */
	if (color < 0x8000) {
		/* use given color but shift it left by 1 */
		color = color << 1;
#ifdef DEBUG_COLOR
		printf("using given color, color is now 0x%04x\n", color);
#endif
	} else if ((color & 0xff) < 0x80) {
		gamecolor2gl_index(red, green, blue, color & 0xf);
		return;
	} else {
		/* use given index from pre-defined palette */
		index = color & 0x7e;
		palette = mercenary_palette_predefined();
		color = m68k_read_memory_16(palette + index);
#ifdef DEBUG_COLOR
		printf("offset (%d) from pre-defined palette (at 0x%x) given, color is now 0x%04x\n", index, palette, color);
#endif
		/* now use that color info parse again (hopefully it does not contain a "pre-defined palette" again and again! */
		if (nesting++ == 8) {
			print_info("Color lookup from pre-defined palette is nesting too much, please fix!\n");
			return;
		}
		goto again;
	}
	*red = (double)((color >> 8) & 0xf) / 15.0;
	*green = (double)((color >> 4) & 0xf) / 15.0;
	*blue = (double)(color & 0xf) / 15.0;
}

static int32_t wrap_int28(int32_t value)
{
	value <<= 4;
	value >>= 4;
	return value;
}

static void store_coord(const char __attribute__((unused)) *what, uint32_t vertex, int32_t x, int32_t y, int32_t z, double scale)
{
	if ((vertex & 3)) {
		print_info("Vertex %d is not a multiple of four!\n", vertex);
		return;
	}
	/* create new vertices item, if there was no vertex before, or if the vertex has different offet */
	if (vertex < 0x100) {
		if (!render_item || render_item->type != RENDER_ITEM_VERTICES_0) {
			render_item_add(RENDER_ITEM_VERTICES_0);
			/* copy vertices that have been captured already */
			if (render_item_vertices_0)
				memcpy(&render_item->u.vertices, &render_item_vertices_0->u.vertices, sizeof(render_item->u.vertices));
			render_item_vertices_0 = render_item;
		}
	} else
	if (vertex < 0x200) {
		if (!render_item || render_item->type != RENDER_ITEM_VERTICES_1) {
			render_item_add(RENDER_ITEM_VERTICES_1);
			/* copy vertices that have been captured already */
			if (render_item_vertices_1)
				memcpy(&render_item->u.vertices, &render_item_vertices_1->u.vertices, sizeof(render_item->u.vertices));
			render_item_vertices_1 = render_item;
		}
		vertex -= 0x100;
	} else
	if (vertex < 0x300) {
		if (!render_item || render_item->type != RENDER_ITEM_VERTICES_2) {
			render_item_add(RENDER_ITEM_VERTICES_2);
			/* copy vertices that have been captured already */
			if (render_item_vertices_2)
				memcpy(&render_item->u.vertices, &render_item_vertices_2->u.vertices, sizeof(render_item->u.vertices));
			render_item_vertices_2 = render_item;
		}
		vertex -= 0x200;
	} else {
		print_info("Vertex %d exceeds maximum vertex number, please fix!\n", vertex);
		return;
	}
	vertex >>= 2;
#ifdef DEBUG_VERTEX
	printf("storing %s coordinates: vertex=%d, x=%d, y=%d, z=%d\n", what, vertex, x, y, z);
#endif
	/* use absolute position */
	render_item->u.vertices.x[vertex] = (double)x * scale + motion_new.position_east;
	render_item->u.vertices.y[vertex] = (double)y * scale + motion_new.position_height;
	render_item->u.vertices.z[vertex] = (double)z * scale + motion_new.position_north;
}

static int use_coord(const char __attribute__((unused)) *what, uint32_t vertex, double *x, double *y, double *z, int fix);

static void store_planets_coord(const char __attribute__((unused)) *what, uint32_t vertex, int32_t x, int32_t y, int32_t z)
{
	if ((vertex & 3)) {
		print_info("Vertex %d is not a multiple of four!\n", vertex);
		return;
	}
	/* create new vertices item, if there was no vertex before, or if the vertex has different offet */
	if (vertex >= MAX_VERTEX) {
		print_info("Vertex %d exceeds maximum vertex number %d!\n", vertex, MAX_INTERIOR_VERTEX);
		return;
	}
	if (!render_item || render_item->type != RENDER_ITEM_VERTICES_0) {
		render_item_add(RENDER_ITEM_VERTICES_0);
		/* copy vertices that have been captured already */
		if (render_item_vertices_0)
			memcpy(&render_item->u.vertices, &render_item_vertices_0->u.vertices, sizeof(render_item->u.vertices));
		render_item_vertices_0 = render_item;
	}
	vertex >>= 2;
#ifdef DEBUG_VERTEX
	printf("storing %s coordinates: vertex=%d, x=%d, y=%d, z=%d\n", what, vertex, x, y, z);
#endif
	render_item->u.vertices.x[vertex] = x;
	render_item->u.vertices.y[vertex] = y;
	render_item->u.vertices.z[vertex] = z;
}

static void store_interior_coord(const char __attribute__((unused)) *what, uint32_t vertex, int32_t x, int32_t y1, int32_t y2, int32_t y3, int32_t y4, int32_t z)
{
	if ((vertex & 3)) {
		print_info("Vertex is not a multiple of four!\n");
		return;
	}
	if (vertex >= MAX_INTERIOR_VERTEX) {
		print_info("Vertex %d exceeds maximum vertex number %d!\n", vertex, MAX_INTERIOR_VERTEX);
		return;
	}
	if (!render_item || render_item->type != RENDER_ITEM_VERTICES_INTERIOR)
		render_item_add(RENDER_ITEM_VERTICES_INTERIOR);
	vertex >>= 2;
#ifdef DEBUG_VERTEX
	printf("storing %s coordinates: vertex=%d, x=%d, y=%d;%d;%d;%d, z=%d\n", what, vertex, x, y1, y2, y3, y4, z);
#endif
	/* use absolute position */
	x += motion_new.position_east;
	y1 += motion_new.position_height;
	y2 += motion_new.position_height;
	y3 += motion_new.position_height;
	y4 += motion_new.position_height;
	z += motion_new.position_north;
	render_item->u.vertices_interior.x[vertex] = (double)x;
	render_item->u.vertices_interior.y[0] = (double)y1;
	render_item->u.vertices_interior.y[1] = (double)y2;
	render_item->u.vertices_interior.y[2] = (double)y3;
	render_item->u.vertices_interior.y[3] = (double)y4;
	render_item->u.vertices_interior.z[vertex] = (double)z;
	render_item->u.vertices_interior.set[vertex] = 1;
}

static void rotate_coordinate(double roll, double pitch, double yaw, double *x, double *y, double *z)
{
	double out_x, out_y, out_z;

	/* rotate yaw (German: Gier, turn view to the right) */
	if (yaw) {
		out_z = (*z) * cos(yaw) - (*x) * sin(yaw);
		out_x = (*z) * sin(yaw) + (*x) * cos(yaw);
		*z = out_z;
		*x = out_x;
	}
	/* rotate pitch (German: Nick, turn head down) */
	if (pitch) {
		out_y = (*y) * cos(pitch) - (*z) * sin(pitch);
		out_z = (*y) * sin(pitch) + (*z) * cos(pitch);
		*y = out_y;
		*z = out_z;
	}
	/* rotate roll (tilt head to the right) */
	if (roll) {
		out_x = (*x) * cos(roll) - (*y) * sin(roll);
		out_y = (*x) * sin(roll) + (*y) * cos(roll);
		*x = out_x;
		*y = out_y;
	}
}

/* clear screen color (sky / universe) */
static void clear_screen(int index)
{
#ifdef DEBUG_VERTEX
	printf("clear screen:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_SKY);

	/* set color */
	gamecolor2gl_index(&render_item->u.ground.red, &render_item->u.ground.green, &render_item->u.ground.blue, 8);

	/* store for later use after planets have been rendered */
	ground_index = index;
}

/* ground color */
static void draw_ground(void)
{
	/* no ground in space :) */
	if (ground_index < 0)
		return;

#ifdef DEBUG_VERTEX
	printf("add ground plane:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_GROUND);

	/* set color */
	gamecolor2gl_index(&render_item->u.ground.red, &render_item->u.ground.green, &render_item->u.ground.blue, ground_index);
}

/* object info */
static void info_object(int moving)
{
#ifdef DEBUG_VERTEX
	printf("add object's info:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_OBJECT_INFO);

	/* add info */
	render_item->u.info.moving = moving;
	if (moving)
		mercenary_get_object_info(&render_item->u.info.id, &render_item->u.info.east, &render_item->u.info.height, &render_item->u.info.north);
	render_item_object_info = render_item;
}

/* coordinates ready for an object */
static void coord_object(void)
{
	int32_t x, y, z;

	x = (int16_t)REG_D[3];
	x += (int32_t)REG_A[1];
	y = (int16_t)REG_D[4];
	y += (int32_t)REG_A[2];
	z = (int16_t)REG_D[5];
	z += (int32_t)REG_A[3];
	store_coord("object", REG_A[0], x, y, z, 1.0);
}

/* polygon of object */
static void poly_object(int mercenary)
{
	uint32_t vertex_address = REG_A[0], va;
	uint32_t vertex;
	double x, y, z;
	int i, rc;

#ifdef DEBUG_VERTEX
	printf("add object's polygon:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_OBJECT_POLYGON);

	/* set color */
	if (mercenary == 3)
		gamecolor2gl(&render_item->u.polygon.red, &render_item->u.polygon.green, &render_item->u.polygon.blue, REG_D[0]);
	else {
		uint16_t color;
		color = m68k_read_memory_8(vertex_address++) << 8;
		color |= m68k_read_memory_8(vertex_address++);
		gamecolor2gl(&render_item->u.polygon.red, &render_item->u.polygon.green, &render_item->u.polygon.blue, color);
	}

	/* the vertex list is zero-terminated */
	va = vertex_address;
	for (i = 0; i < MAX_POLYGON; i++) {
		vertex = m68k_read_memory_8(va++);
		if (vertex == 0 && i)
			break;
		render_item->u.polygon.vertex[i] = vertex;
	}
	render_item->u.polygon.vertices = i;

	/* apply only to not fixed objects (a fixed may be taxi interior) */
	if (render_item_object_info && render_item_object_info->u.info.moving) {
		/* allocate shadow item */
		render_item_add(RENDER_ITEM_SHADOW_OBJECT_POLYGON);
		render_item->u.shadow_polygon.x_pos = -motion_new.position_east;
		render_item->u.shadow_polygon.y_pos = -motion_new.position_height;
		render_item->u.shadow_polygon.z_pos = -motion_new.position_north;

		/* the vertex list is zero-terminated */
		va = vertex_address;
		for (i = 0; i < MAX_POLYGON; i++) {
			vertex = m68k_read_memory_8(va++);
			if (vertex == 0 && i)
				break;
			rc = use_coord("object", vertex, &x, &y, &z, 1);
			if (rc < 0)
				break;
			render_item->u.shadow_polygon.x[i] = x + motion_new.position_east;
			render_item->u.shadow_polygon.y[i] = y + motion_new.position_height;
			render_item->u.shadow_polygon.z[i] = z + motion_new.position_north;
		}
		render_item->u.shadow_polygon.vertices = i;
	}
}

/* line of object */
static void line_object(void)
{
	uint32_t vertex_address = REG_A[0], va;
	uint32_t vertex;
	double x, y, z;
	int rc;

#ifdef DEBUG_VERTEX
	printf("add object's line:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_OBJECT_LINE);

	/* set color */
	gamecolor2gl(&render_item->u.line.red, &render_item->u.line.green, &render_item->u.line.blue, REG_D[0]);

	/* two vertices */
	va = vertex_address;
	vertex = m68k_read_memory_8(va++);
	render_item->u.line.vertex[0] = vertex;
	vertex = m68k_read_memory_8(va++);
	render_item->u.line.vertex[1] = vertex;

	/* allocate shadow item */
	render_item_add(RENDER_ITEM_SHADOW_OBJECT_LINE);
	render_item->u.shadow_line.x_pos = -motion_new.position_east;
	render_item->u.shadow_line.y_pos = -motion_new.position_height;
	render_item->u.shadow_line.z_pos = -motion_new.position_north;

	/* apply only to not fixed objects (a fixed may be taxi interior) */
	if (render_item_object_info && render_item_object_info->u.info.moving) {
		va = vertex_address;
		vertex = m68k_read_memory_8(va++);
		rc = use_coord("object", vertex, &x, &y, &z, 1);
		if (rc < 0)
			x=y=z=0;
		render_item->u.shadow_line.x[0] = x + motion_new.position_east;
		render_item->u.shadow_line.y[0] = y + motion_new.position_height;
		render_item->u.shadow_line.z[0] = z + motion_new.position_north;
		vertex = m68k_read_memory_8(va++);
		rc = use_coord("object", vertex, &x, &y, &z, 1);
		if (rc < 0)
			x=y=z=0;
		render_item->u.shadow_line.x[1] = x + motion_new.position_east;
		render_item->u.shadow_line.y[1] = y + motion_new.position_height;
		render_item->u.shadow_line.z[1] = z + motion_new.position_north;
	}
}

/* coordinates ready for a beacon */
static void coord_beacon(void)
{
	int32_t x, y, z;

	/* only 28 bits seem to be a correct signed int value */
	x = (int32_t)(REG_D[3] << 4) / 16;
	y = (int32_t)(REG_D[4] << 4) / 16;
	z = (int32_t)(REG_D[5] << 4) / 16;
	store_coord("beacon", 0, x, y, z, 1.0);
}

/* point of beacon */
static void point_beacon(void)
{
#ifdef DEBUG_VERTEX
	printf("add beacon's point:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_BEACON_POINT);

	/* set color */
	gamecolor2gl(&render_item->u.point.red, &render_item->u.point.green, &render_item->u.point.blue, REG_D[2]);

	/* one vertex */
	render_item->u.point.vertex = 0;
}

/* coordinates ready for a building (exterior) */
static void coord_building_exterior(void)
{
	int32_t x, y, z;

	x = (int32_t)REG_D[3];
	y = (int32_t)REG_D[4];
	z = (int32_t)REG_D[5];
	store_coord("building exterior", REG_A[0], x, y, z, 1.0);
}

/* polygon of building (exterior) */
static void poly_building_exterior(void)
{
	uint16_t color;
	uint32_t vertex_address = REG_A[0];
	uint32_t vertex;
	int i;

#ifdef DEBUG_VERTEX
	printf("add building's polygon:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_BUILDING_EXTERIOR_POLYGON);

	/* set color */
	color = m68k_read_memory_8(vertex_address++) << 8;
	color |= m68k_read_memory_8(vertex_address++);
	gamecolor2gl(&render_item->u.polygon.red, &render_item->u.polygon.green, &render_item->u.polygon.blue, color);

	/* the vertex list is zero-terminated */
	for (i = 0; i < MAX_POLYGON; i++) {
		vertex = m68k_read_memory_8(vertex_address++);
		if (vertex == 0 && i)
			break;
		render_item->u.polygon.vertex[i] = vertex | 0x100;
	}
	render_item->u.polygon.vertices = i;
}

/* line of building (exterior) */
static void line_building_exterior(void)
{
	uint32_t vertex_address = REG_A[0];
	uint32_t vertex;

#ifdef DEBUG_VERTEX
	printf("add building's line:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_BUILDING_EXTERIOR_LINE);

	/* set color */
	gamecolor2gl(&render_item->u.line.red, &render_item->u.line.green, &render_item->u.line.blue, REG_D[0]);

	/* two vertices */
	vertex = m68k_read_memory_8(vertex_address++);
	render_item->u.line.vertex[0] = vertex | 0x100;
	vertex = m68k_read_memory_8(vertex_address++);
	render_item->u.line.vertex[1] = vertex | 0x100;
}

/* coordinates ready for a building (interior) */
static void coord_building_interior(void)
{
	int16_t east, north;
	int32_t height1, height2, height3, height4;

	mercenary_coord_building_interior(&east, &height1, &height2, &height3, &height4, &north);
	store_interior_coord("interior", REG_A[0], east, height1, height2, height3, height4, north);
}

/* polygon of building (interior) */
static void poly_building_interior1to4(int level)
{
	uint16_t color;
	uint32_t vertex;
	int i;

#ifdef DEBUG_VERTEX
	printf("add roof/floor's polygon at level %d:\n", level);
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_BUILDING_INTERIOR_1TO4);

	/* set color */
	color = m68k_read_memory_8(REG_A[0]) << 8;
	color |= m68k_read_memory_8(REG_A[0] + 1);
	gamecolor2gl(&render_item->u.interior14.red, &render_item->u.interior14.green, &render_item->u.interior14.blue, color);

	/* four vertices, one level */
	for (i = 0; i < 4; i++) {
		vertex = REG_A[(2 + i)];
		render_item->u.interior14.vertex[i] = vertex;
	}
	render_item->u.interior14.level = level;
}

/* polygon of building (interior) */
static void poly_building_interior5to6(int level12, int level34)
{
	uint16_t color;

#ifdef DEBUG_VERTEX
	printf("add polygon above/below window/door at level %d/%d:\n", level12, level34);
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_BUILDING_INTERIOR_5TO6);

	/* set color */
	color = m68k_read_memory_8(REG_A[0]) << 8;
	color |= m68k_read_memory_8(REG_A[0] + 1);
	gamecolor2gl(&render_item->u.interior56.red, &render_item->u.interior56.green, &render_item->u.interior56.blue, color);

	/* two vertices, two levels */
	render_item->u.interior56.vertex14 = REG_A[2];
	render_item->u.interior56.vertex23 = REG_A[3];
	render_item->u.interior56.level12 = level12;
	render_item->u.interior56.level34 = level34;
}

/* wall part of a building */
static void wall_building(void)
{
#ifdef DEBUG_VERTEX
	printf("add wall:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_BUILDING_INTERIOR_WALL);

	/* set color */
	gamecolor2gl(&render_item->u.interior56.red, &render_item->u.interior56.green, &render_item->u.interior56.blue, REG_D[3]);

	/* two vertices, two levels */
	render_item->u.interior56.vertex14 = REG_A[1];
	render_item->u.interior56.vertex23 = REG_A[2];
	/* get top level according to bit 12 in D3 */
	render_item->u.interior56.level12 = (REG_D[3] & (1 << 12)) ? 3 : 2;
	render_item->u.interior56.level34 = 1;
}

/* coordinates ready for comet tail */
static void coord_comet(void)
{
	int32_t x, y, z;

	x = (int32_t)REG_D[3];
	y = (int32_t)REG_D[4];
	z = (int32_t)REG_D[5];
	store_planets_coord("comet tail", REG_A[0], x, y, z);
}

/* polygon of comet tail */
static void poly_comet(void)
{
	uint16_t color;
	uint32_t vertex_address = REG_A[0];
	uint32_t vertex;
	int i;

#ifdef DEBUG_VERTEX
	printf("add comet's polygon:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_COMET_POLYGON);

	/* set color */
	color = m68k_read_memory_8(vertex_address++) << 8;
	color |= m68k_read_memory_8(vertex_address++);
	gamecolor2gl(&render_item->u.polygon.red, &render_item->u.polygon.green, &render_item->u.polygon.blue, color);

	/* the vertex list is zero-terminated */
	for (i = 0; i < MAX_POLYGON; i++) {
		vertex = m68k_read_memory_8(vertex_address++);
		if (vertex == 0 && i)
			break;
		render_item->u.polygon.vertex[i] = vertex;
	}
	render_item->u.polygon.vertices = i;
}

/* coordinates ready for lines of a road / ground surface */
static void coord_line_road(void)
{
	int32_t x, y, z;

	x = REG_D[3];
	y = -motion_new.position_height;
	z = REG_D[5];
	store_coord("road", REG_A[0], x, y, z, 1.0);
}

/* line of road */
static void line_road(void)
{
	uint32_t vertex;

#ifdef DEBUG_VERTEX
	printf("add road's line:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_ROAD_LINE);

	/* set color */
	gamecolor2gl_index(&render_item->u.line.red, &render_item->u.line.green, &render_item->u.line.blue, mercenary_street_color_index());

	/* two vertices */
	vertex = REG_A[1];
	render_item->u.line.vertex[0] = vertex;
	vertex = REG_A[2];
	render_item->u.line.vertex[1] = vertex;
}

/* coordinates ready for polygons of a road / ground surface */
static void coord_poly_road(void)
{
	int32_t x, y, z;

	x = m68k_read_memory_32(320 + REG_A[0]);
	x -= REG_A[1];
	/* the A2 is already converted to game's coordinate, so we use the memory location DS_0+0x1DBA (m3) instead */
	y = -motion_new.position_height;
	z = m68k_read_memory_32(576 + REG_A[0]);
	z -= REG_A[3];
	store_coord("road/place", REG_A[0], x, y, z, 1.0);
}

/* polygon of road */
static void poly_road()
{
	uint16_t color;
	uint32_t vertex_address = REG_A[0];
	uint32_t vertex;
	int i;

#ifdef DEBUG_VERTEX
	printf("add road/place's polygon:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_ROAD_POLYGON);

	/* set color */
	color = m68k_read_memory_8(vertex_address++) << 8;
	color |= m68k_read_memory_8(vertex_address++);
	gamecolor2gl(&render_item->u.polygon.red, &render_item->u.polygon.green, &render_item->u.polygon.blue, color);

	/* the vertex list is zero-terminated */
	for (i = 0; i < MAX_POLYGON; i++) {
		vertex = m68k_read_memory_8(vertex_address++);
		if (vertex == 0 && i)
			break;
		render_item->u.polygon.vertex[i] = vertex;
	}
	render_item->u.polygon.vertices = i;
}

/* coordinates ready for tags */
static void coord_tags(void)
{
	int32_t x, y, z;

	x = (int16_t)REG_D[3];
	x += (int32_t)REG_A[1];
	y = (int16_t)REG_D[4];
	y += (int32_t)REG_A[2];
	z = (int16_t)REG_D[5];
	z += (int32_t)REG_A[3];
	store_coord("tags", REG_A[0], x, y, z, 1.0);
}

/* coordinates ready for large tags */
static void coord_tags2(void)
{
	int32_t x, y, z;

	x = (int16_t)REG_D[3];
	x += 2 * (int32_t)REG_A[1];
	y = (int16_t)REG_D[4];
	y += 2 * (int32_t)REG_A[2];
	z = (int16_t)REG_D[5];
	z += 2 * (int32_t)REG_A[3];
	/* note that large tags have double distance, so the resolution is vitrually doubled.
	 * since we use interpolation and VR, we need to scale the vertex back to normal distance.
	 */
	store_coord("large tags", REG_A[0], x, y, z, 0.5);
}

/* line of tag */
static void line_tags(int last_color)
{
	uint32_t vertex_address = REG_A[0];
	uint32_t vertex;

#ifdef DEBUG_VERTEX
	printf("add tag's line:\n");
#endif

	/* allocate render item */
	render_item_add((tag_is_object) ? RENDER_ITEM_TAG_LINE_OBJECT : RENDER_ITEM_TAG_LINE_OTHER);

	/* set color */
	if (!last_color)
		gamecolor2gl(&render_item->u.line.red, &render_item->u.line.green, &render_item->u.line.blue, REG_D[0]);
	else
		gamecolor2gl_index(&render_item->u.line.red, &render_item->u.line.green, &render_item->u.line.blue, mercenary_line_tags_index());

	/* two vertices */
	vertex = m68k_read_memory_8(vertex_address++);
	render_item->u.line.vertex[0] = vertex | 0x200;
	vertex = m68k_read_memory_8(vertex_address++);
	render_item->u.line.vertex[1] = vertex | 0x200;
}

/* polygon of tags */
static void poly_tags(int last_color)
{
	uint32_t vertex_address = REG_A[0];
	uint32_t vertex;
	int i;

#ifdef DEBUG_VERTEX
	printf("add tag's polygon:\n");
#endif

	/* allocate render item */
	render_item_add((tag_is_object) ? RENDER_ITEM_TAG_POLYGON_OBJECT : RENDER_ITEM_TAG_POLYGON_OTHER);

	/* set color */
	if (!last_color)
		gamecolor2gl(&render_item->u.polygon.red, &render_item->u.polygon.green, &render_item->u.polygon.blue, REG_D[0]);
	else
		gamecolor2gl(&render_item->u.polygon.red, &render_item->u.polygon.green, &render_item->u.polygon.blue, mercenary_poly_tags_color());
	/* the vertex list is zero-terminated */
	for (i = 0; i < MAX_POLYGON; i++) {
		vertex = m68k_read_memory_8(vertex_address++);
		if (vertex == 0 && i)
			break;
		render_item->u.polygon.vertex[i] = vertex | 0x200;
	}
	render_item->u.polygon.vertices = i;
}

/* coordinates ready for planet */
static void coord_planet(void)
{
	int32_t x, y, z;

	x = (int32_t)REG_D[3];
	y = (int32_t)REG_D[4];
	z = (int32_t)REG_D[5];
	store_planets_coord("planet", REG_A[0], x, y, z);
}

/* planet */
static void draw_planet(int comet)
{
	uint32_t vertex;
	uint32_t scale_index;
	double scale1, scale2;

	vertex = REG_A[0];

	/* fixing (not noticable) bug in game: don't render comet twice */
	if (!comet && vertex == 116)
		return;

#ifdef DEBUG_VERTEX
	printf("add planet/comet/sun: vertex=%d color=%08x\n", vertex, color);
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_PLANET);

	/* set color */
	if (comet) {
		/* make comet black on front side and bright on back */
		gamecolor2gl(&render_item->u.planet.front_red, &render_item->u.planet.front_green, &render_item->u.planet.front_blue, 0x000);
		gamecolor2gl(&render_item->u.planet.back_red, &render_item->u.planet.back_green, &render_item->u.planet.back_blue, 0x777);

	} else {
		gamecolor2gl(&render_item->u.planet.front_red, &render_item->u.planet.front_green, &render_item->u.planet.front_blue, REG_D[0]);
		/* use background color for dark side */
		gamecolor2gl(&render_item->u.planet.back_red, &render_item->u.planet.back_green, &render_item->u.planet.back_blue, mercenary_background_index() | 0x8000);
	}

	/* set vertex */
	render_item->u.planet.vertex = vertex;

	/* I REVERSED THE F*CKING PLANET SIZE, took me half a day!
	 * the long word 21584(A0) contains two scales
	 * the lower word contains the scale between 4096 .. 8191, this is 1.0 .. 2.0
	 * the upper word defines how much this scale is shifted to the left.
	 */
	scale_index = mercenary_planet_scale_index();
	scale1 = (double)(m68k_read_memory_16(scale_index + 2 + vertex) & 0x1fff) / 4096.0;
	scale2 = (double)(1 << (uint16_t)m68k_read_memory_16(scale_index + vertex));
	render_item->u.planet.size = scale1 * scale2 / 256.0;
}

/* stars */
static void draw_stars(int16_t v_offset, int tilt, int above_zenith)
{
#ifdef DEBUG_VERTEX
	printf("add stars\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_STARS);

	/* vertical offset */
	render_item->u.stars.v_offset = v_offset;

	/* if we fly over the planet, we have tilt, so we get the calculated tilt value from game */
	render_item->u.stars.tilt = tilt;
	if (tilt)
		render_item->u.stars.tilt_value = (int32_t)REG_A[4];

	/* stars above zenith */
	render_item->u.stars.above_zenith = above_zenith;
}

/* stars of interstellar flight */
static void draw_stars_interstellar(void)
{
	int i, count;
	uint32_t table;

#ifdef DEBUG_VERTEX
	printf("add interstellar stars\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_INTERSTELLAR_STARS);

	table = REG_A[0];
	count = REG_D[5] + 1;
	if (count > MAX_INTERSTARS) {
		print_info("Expecting maximum of %d stars here, plese fix!\n", MAX_INTERSTARS);
		return;
	}
	for (i = 0; i < count; i++) {
		table += 12;
		/* get color */
		render_item->u.interstars.color[i] = (m68k_read_memory_16(table) >> 5) & 0xf;
		table += 2;
		render_item->u.interstars.x[i] = m68k_read_memory_16(table);
		table += 2;
		render_item->u.interstars.y[i] = m68k_read_memory_16(table);
		table += 2;
	}
	render_item->u.interstars.count = count;
}

/* sun of interstellar flight (center dot) */
static void draw_sun_interstellar(void)
{
#ifdef DEBUG_VERTEX
	printf("add interstellar sun\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_INTERSTELLAR_SUN);
}

/* coordinates ready for polygons of islands */
static void coord_islands(void)
{
	int32_t x, y, z;

	x = ((int16_t)m68k_read_memory_16(REG_A[4] - 2) * 65536);
	x += (int32_t)REG_A[1];
	/* the A2 is already converted to game's coordinate, so we use the memory location DS_0+0x1DBA instead */
	y = -motion_new.position_height;
	z = ((int16_t)m68k_read_memory_16(REG_A[4]) * 65536);
	z += (int32_t)REG_A[3];
	store_coord("island", REG_A[0], x, y, z, 1.0);
}

/* polygon of island */
static void poly_island()
{
	uint16_t color;
	uint32_t vertex_address = REG_A[0];
	uint32_t vertex;
	int i;

#ifdef DEBUG_VERTEX
	printf("add island:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_ISLAND_POLYGON);

	/* set color */
	color = m68k_read_memory_8(vertex_address++) << 8;
	color |= m68k_read_memory_8(vertex_address++);
	gamecolor2gl(&render_item->u.polygon.red, &render_item->u.polygon.green, &render_item->u.polygon.blue, color);

	/* the vertex list is zero-terminated */
	i = 0;
	while (i < MAX_POLYGON) {
		vertex = m68k_read_memory_8(vertex_address++);
		if (vertex == 0 && i)
			break;
		/* skip mysterious points when rendering island */
		if (vertex >= 0xf0)
			continue;
		render_item->u.polygon.vertex[i] = vertex;
		i++;
	}
	render_item->u.polygon.vertices = i;
}

/* sights */
static void draw_sights(void)
{
#ifdef DEBUG_VERTEX
	printf("add sights:\n");
#endif

	/* allocate render item */
	render_item_add(RENDER_ITEM_SIGHTS);
}

static int32_t coord_explosion_x, coord_explosion_y, coord_explosion_z;

static void coord_explosion(void)
{
	coord_explosion_x = REG_D[3];
	coord_explosion_y = REG_D[4];
	coord_explosion_z = REG_D[5];
}

static void draw_explosion(void)
{
	uint16_t color;
	double scale1, scale2;

#ifdef DEBUG_VERTEX
	printf("add explosion:\n");
#endif

	/* allocate render item */
	if (!render_item || render_item->type != RENDER_ITEM_EXPLOSION) {
		render_item_add(RENDER_ITEM_EXPLOSION);
		/* get color from render palette */
		color = (m68k_read_memory_16(mercenary_palette_view() + 13*2) & 0xfff) >> 1;
		gamecolor2gl(&render_item->u.explosion.red, &render_item->u.explosion.green, &render_item->u.explosion.blue, color);
		render_item->u.explosion.count = 0;
	}
	if (render_item->u.explosion.count == MAX_EXPLOSION)
		return;

	/* the D1 register contains two scales
	 * the lower word contains the scale between 4096 .. 8191, this is 1.0 .. 2.0
	 * the upper word defines how much this scale is shifted to the left.
	 */
	if ((int16_t)(REG_D[1] >> 16) < 0) {
		scale1 = (double)(REG_D[1] & 0x1fff) / 4096.0;
		scale2 = (double)(256 >> -(int16_t)(REG_D[1] >> 16)) / 256.0;
	} else
	if ((int16_t)(REG_D[1] >> 16) <= 10) {
		scale1 = (double)(REG_D[1] & 0x1fff) / 4096.0;
		scale2 = (double)(1 << (int16_t)(REG_D[1] >> 16));
	} else {
		scale1 = 1.0;
		scale2 = 1 << 10;
	}
	render_item->u.explosion.size[render_item->u.explosion.count] = scale1 * scale2;
	render_item->u.explosion.x[render_item->u.explosion.count] = coord_explosion_x;
	render_item->u.explosion.y[render_item->u.explosion.count] = coord_explosion_y;
	render_item->u.explosion.z[render_item->u.explosion.count] = coord_explosion_z;
	render_item->u.explosion.count++;
}

/* get building data out of game data, so we use it to render shadow */
static void draw_shadow_building(void)
{
	int32_t scale_x, scale_y, scale_z;
	int32_t x, y, z;
	uint16_t anim_flag, anim_x, anim_y, anim_z, anim_phase;
	uint32_t a0, a0s, a5, a6, a4;
	uint16_t s;
	int16_t d3, d4, d5;
	int32_t d3x, d4x, d5x;
	int d1index;
	int line_from, line_to, point;
	int i;
	struct render_item_vertices vertex;

	mercenary_get_building_exterior_info(&x, &z, &scale_x, &scale_y, &scale_z, &anim_flag, &anim_x, &anim_y, &anim_z, &anim_phase, &a4, &a0, &a5);
	x -= motion_new.position_east;
	y = -motion_new.position_height;
	z -= motion_new.position_north;

	/* list of all vetices */
	for (i = 1; i < 64; i++) {
		d3 = m68k_read_memory_16(a4);
		a4 += 2;
		if (d3 <= (int16_t)0x8001)
			break;
		d4 = m68k_read_memory_16(a4);
		a4 += 2;
		d5 = m68k_read_memory_16(a4);
		a4 += 2;
		if (d4 < 0) {
			d4 = ~d4;
			/* scale */
			d3x = d3 * scale_x;
			d4x = d4 * scale_y;
			d5x = d5 * scale_z;
			/* angle */
			double angle = (double)anim_phase / 65536.0 * 2.0 * M_PI;
			/* animate around what axis ? */
			if ((int16_t)anim_flag > 0) {
				/* rotate around z */
				d3 = (d3x & 0xffff) - anim_x;
				d4 = (d4x & 0xffff) - anim_y;
				if (angle) {
					double cosd4 = (double) d4 * cos(angle);
					double sind3 = (double) d3 * sin(angle);
					double cosd3 = (double) d3 * cos(angle);
					double sind4 = (double) d4 * sin(angle);
					d3 = cosd3 - sind4;
					d4 = cosd4 + sind3;
				}
				d3x = d3 + anim_x;
				d4x = d4 + anim_y;
			}
			if ((int16_t)anim_flag == 0) {
				/* rotate around y */
				d3 = (d3x & 0xffff) - anim_x;
				d5 = (d5x & 0xffff) - anim_z;
				if (angle) {
					double cosd3 = (double) d3 * cos(angle);
					double sind5 = (double) d5 * sin(angle);
					double cosd5 = (double) d5 * cos(angle);
					double sind3 = (double) d3 * sin(angle);
					d3 = cosd3 + sind5;
					d5 = cosd5 - sind3;
				}
				d3x = d3 + anim_x;
				d5x = d5 + anim_z;
			}
			if ((int16_t)anim_flag < 0) {
				/* rotate around x */
				d4 = (d4x & 0xffff) - anim_y;
				d5 = (d5x & 0xffff) - anim_z;
				if (angle) {
					double cosd5 = (double) d5 * cos(angle);
					double sind4 = (double) d4 * sin(angle);
					double cosd4 = (double) d4 * cos(angle);
					double sind5 = (double) d5 * sin(angle);
					d4 = cosd4 + sind5;
					d5 = cosd5 - sind4;
				}
				d4x = d4 + anim_y;
				d5x = d5 + anim_z;
			}
		} else {
			/* scale */
			d3x = d3 * scale_x;
			d4x = d4 * scale_y;
			d5x = d5 * scale_z;
		}
		vertex.x[i] = d3x;
		vertex.y[i] = d4x;
		vertex.z[i] = d5x;
	}

	/* get all data sets */
	do {
		s = m68k_read_memory_16(a0);
		a0 += 2;
		if (s) {
			/* skip over distance cube */
			a0 += 4 * 6;
		}
		/* get address pointer to data set */
		a0s = m68k_read_memory_32(a0);
		a0 += 4;

		/* walk through primitives (polygons and lines) */
		while ((d1index = m68k_read_memory_8(a0s++))) {
			/* pointer to primitive */
			a6 = m68k_read_memory_32(a5 + d1index*4);
			if ((m68k_read_memory_8(a6) & 0x40) == 0) {
				/* polygon */
				a6 += 2;
#ifdef DEBUG_VERTEX
				printf("add object's shadow polygon:\n");
#endif
				/* allocate render item */
				render_item_add(RENDER_ITEM_SHADOW_BUILDING_POLYGON);
				render_item->u.shadow_polygon.x_pos = x;
				render_item->u.shadow_polygon.y_pos = y;
				render_item->u.shadow_polygon.z_pos = z;
				i = 0;
				while ((point = m68k_read_memory_8(a6++) >> 2)) {
					render_item->u.shadow_polygon.x[i] = vertex.x[point];
					render_item->u.shadow_polygon.y[i] = vertex.y[point];
					render_item->u.shadow_polygon.z[i] = vertex.z[point];
					i++;
				}
				render_item->u.shadow_polygon.vertices = i;
			} else if ((m68k_read_memory_16(a6) & 0x2000) == 0) {
				/* line */
				line_from = m68k_read_memory_8(a6 + 2) >> 2;
				line_to = m68k_read_memory_8(a6 + 3) >> 2;
				/* allocate render item */
				render_item_add(RENDER_ITEM_SHADOW_BUILDING_LINE);
				render_item->u.shadow_line.x_pos = x;
				render_item->u.shadow_line.y_pos = y;
				render_item->u.shadow_line.z_pos = z;
				render_item->u.shadow_line.x[0] = vertex.x[line_from];
				render_item->u.shadow_line.y[0] = vertex.y[line_from];
				render_item->u.shadow_line.z[0] = vertex.z[line_from];
				render_item->u.shadow_line.x[1] = vertex.x[line_to];
				render_item->u.shadow_line.y[1] = vertex.y[line_to];
				render_item->u.shadow_line.z[1] = vertex.z[line_to];
			}
		}
	} while (s);
}

/* stop event from CPU received */
void render_capture_event(int event)
{
	switch (event) {
	case STOP_AT_CLEAR_SCREEN1:
		clear_screen(16); /* color 16 is raster split */
		/* in case of screen clearing on the ground, there is no roll */
		motion_new.orientation_roll = 0;
		break;
	case STOP_AT_CLEAR_SCREEN2:
		clear_screen(15);
		break;
	case STOP_AT_CLEAR_SCREEN3:
		clear_screen(-1); /* no ground (in universe) */
		break;
	case STOP_AT_DRAW_GROUND:
		draw_ground();
		break;
	case STOP_AT_INFO_OBJECT_MOVING:
		info_object(1);
		break;
	case STOP_AT_INFO_OBJECT_FIX:
		info_object(0);
		break;
	case STOP_AT_TAG_IS_OBJECT_1:
		tag_is_object = 1;
		break;
	case STOP_AT_TAG_IS_OBJECT_0:
		tag_is_object = 0;
		break;
	case STOP_AT_COORD_OBJECT:
		coord_object();
		break;
	case STOP_AT_POLY_OBJECT_M3:
		poly_object(3);
		break;
	case STOP_AT_POLY_OBJECT_M2:
		poly_object(2);
		break;
	case STOP_AT_LINE_OBJECT:
		line_object();
		break;
	case STOP_AT_COORD_BEACON:
		coord_beacon();
		break;
	case STOP_AT_POINT_BEACON:
		point_beacon();
		/* note: we may not call the point-renderer, because projected coordinates are invalid */
		mercenary_patch_render(draw_shadows);
		break;
	case STOP_AT_COORD_BUILDING_EXTERIOR:
		coord_building_exterior();
		break;
	case STOP_AT_POLY_BUILDING_EXTERIOR:
		poly_building_exterior();
		break;
	case STOP_AT_LINE_BUILDING_EXTERIOR:
		line_building_exterior();
		break;
	case STOP_AT_COORD_BUILDING_INTERIOR:
		coord_building_interior();
		break;
	case STOP_AT_POLY_BUILDING_INTERIOR1:
		/* floor */
		interior_level12 = 1;
		interior_level34 = 1;
		break;
	case STOP_AT_POLY_BUILDING_INTERIOR2:
		/* ceiling */
		interior_level12 = 2;
		interior_level34 = 2;
		break;
	case STOP_AT_POLY_BUILDING_INTERIOR3:
		/* door/window top */
		interior_level12 = 3;
		interior_level34 = 3;
		break;
	case STOP_AT_POLY_BUILDING_INTERIOR4:
		/* window bottom */
		interior_level12 = 4;
		interior_level34 = 4;
		break;
	case STOP_AT_POLY_BUILDING_INTERIOR5:
		/* door/window top */
		interior_level12 = 2;
		interior_level34 = 3;
		break;
	case STOP_AT_POLY_BUILDING_INTERIOR6:
		/* window bottom */
		interior_level12 = 1;
		interior_level34 = 4;
		break;
	case STOP_AT_POLY_BUILDING_INTERIOR1to4:
		/* before we come here, we must already passed the break points above, so we know the level to be rendered */
		if (interior_level12 == 0) {
			print_info("Interior level is not set, please fix!\n");
			break;
		}
		poly_building_interior1to4(interior_level12);
		interior_level12 = 0;
		break;
	case STOP_AT_POLY_BUILDING_INTERIOR5to6:
		/* before we come here, we must already passed the break points above, so we know the level to be rendered */
		if (interior_level12 == 0) {
			print_info("Interior level is not set, please fix!\n");
			break;
		}
		poly_building_interior5to6(interior_level12, interior_level34);
		interior_level12 = 0;
		break;
	case STOP_AT_WALL_BUILDING:
		wall_building();
		break;
	case STOP_AT_COORD_COMET:
		coord_comet();
		break;
	case STOP_AT_MATRIX_COMET:
	case STOP_AT_MATRIX_PLANET:
		/* track the rotation matrix
		 * if we have 0x42c44 matrix, we must add extra rotation to planet.
		 * the rotation will change the view from the planet's surface */
		if (REG_A[4] == 0x42c44) /* same with M2 and M3 */
			motion_new.planet_rotation = 1;
		else
			motion_new.planet_rotation = 0;
		break;
	case STOP_AT_POLY_COMET:
		poly_comet();
		break;
	case STOP_AT_COORD_LINE_ROADS:
		coord_line_road();
		break;
	case STOP_AT_LINE_ROADS:
		line_road();
		break;
	case STOP_AT_COORD_POLY_ROADS:
		coord_poly_road();
		break;
	case STOP_AT_LINE_ROADS_CENTER:
		/* we don't need to render center lines of roads, because there are polygons already
		 * it does not make sense, since OpenGL has much higher resolution.
		 */
		break;
	case STOP_AT_POLY_ROADS:
		poly_road();
		break;
	case STOP_AT_COORD_TAGS:
		coord_tags();
		break;
	case STOP_AT_COORD_TAGS2:
		coord_tags2();
		break;
	case STOP_AT_LINE_TAGS1:
		line_tags(0);
		break;
	case STOP_AT_LINE_TAGS2:
		line_tags(1);
		break;
	case STOP_AT_POLY_TAGS1:
		poly_tags(0);
		break;
	case STOP_AT_POLY_TAGS2:
		poly_tags(1);
		break;
	case STOP_AT_COORD_PLANET:
		coord_planet();
		break;
	case STOP_AT_DRAW_PLANET:
		draw_planet(0);
		break;
	case STOP_AT_DRAW_COMET:
		draw_planet(1);
		break;
	case STOP_AT_DRAW_STARS_SPACE:
		/* render stars with vertical offset 0x1c0, no tilt, not above zenith */
		draw_stars(0x1c0, 0, 0);
		break;
	case STOP_AT_DRAW_STARS_GROUND:
		/* render stars with vertical offset 0x128, no tilt, not above zenith */
		draw_stars(0x128, 0, 0); /* yet it's hex 128! */
		break;
	case STOP_AT_DRAW_STARS_FLYING:
		/* render stars with vertical offset 0x128, with tilt, not above zenith */
		draw_stars(0x128, 1, 0); /* yet it's hex 128! */
		break;
	case STOP_AT_DRAW_STARS_FLYING2:
		/* render stars with vertical offset 0x128, with tilt, and above zenith */
		draw_stars(0x128, 1, 1); /* yet it's hex 128! */
		break;
	case STOP_AT_DRAW_STARS_INTERSTELLAR:
		draw_stars_interstellar();
		break;
	case STOP_AT_DRAW_SUN_INTERSTELLAR:
		draw_sun_interstellar();
		break;
	case STOP_AT_COORD_ISLANDS:
		coord_islands();
		break;
	case STOP_AT_POLY_ISLANDS:
		poly_island();
		break;
	case STOP_AT_LINE_ISLANDS:
		/* this is not used, as i had noticed so far */
		puts("line island");
#warning FIXME: this is used on LACHESIS to improve visibility of stripes on the ground. implement this and compate a version with lines to the version without lines
		break;
	case STOP_AT_DRAW_SIGHTS:
		draw_sights();
		break;
	case STOP_AT_POLY_UKN2:
		puts("poly ukn2");
		break;
	case STOP_AT_COORD_EXPLOSION:
		coord_explosion();
		break;
	case STOP_AT_DRAW_EXPLOSION:
		draw_explosion();
		break;
	case STOP_AT_PATCH_RENDER:
		mercenary_patch_render(draw_shadows);
		break;
	case STOP_AT_SHADOW_BUILDING:
		draw_shadow_building();
		break;
	}
}

/*
 * rendering
 */

static int use_coord(const char __attribute__((unused)) *what, uint32_t vertex, double *x, double *y, double *z, int no_inter)
{
	render_item_t *ri = NULL;

	if ((vertex & 3)) {
		print_info("Vertex %d is not a multiple of four!\n", vertex);
		return -1;
	}
	if (vertex < 0x100) {
		if (!render_item_vertices_0) {
			print_info("Vertices item for vertex %d not yet set!\n", vertex);
			return -1;
		}
		ri = render_item_vertices_0;
	} else
	if (vertex < 0x200) {
		if (!render_item_vertices_1) {
			print_info("Vertices item for vertex %d not yet set!\n", vertex);
			return -1;
		}
		ri = render_item_vertices_1;
		vertex -= 0x100;
	} else
	if (vertex < 0x300) {
		if (!render_item_vertices_2) {
			print_info("Vertices item for vertex %d not yet set!\n", vertex);
			return -1;
		}
		ri = render_item_vertices_2;
		vertex -= 0x200;
	} else {
		print_info("Vertex %d exceeds maximum vertex number %d!\n", vertex, MAX_VERTEX);
		return -1;
	}
	vertex >>= 2;
	/* translate to original position */
	*x = ri->u.vertices.x[vertex] - motion_new.position_east;
	*y = ri->u.vertices.y[vertex] - motion_new.position_height;
	*z = ri->u.vertices.z[vertex] - motion_new.position_north;
	if (!no_inter) {
		/* translate to floating (interpolated) position offset */
		*x -= interpolation.offset_east;
		*y -= interpolation.offset_height;
		*z -= interpolation.offset_north;
	}
#ifdef DEBUG_VERTEX
	printf("using %s coordinates: vertex=%d, x=%.0f, y=%.0f, z=%.0f\n", what, vertex, *x, *y, *z);
#endif

	return 0;
}

static int use_planet_coord(const char __attribute__((unused)) *what, uint32_t vertex, double *x, double *y, double *z)
{
	render_item_t *ri = NULL;

	if ((vertex & 3)) {
		print_info("Vertex %d is not a multiple of four!\n", vertex);
		return -1;
	}
	if (vertex >= MAX_VERTEX) {
		print_info("Vertex %d exceeds maximum vertex number %d!\n", vertex, MAX_VERTEX);
		return -1;
	}
	if (interpolation.planets)
		ri = interpolation.planets;
	else
		ri = render_item_vertices_0;
	if (!ri) {
		print_info("Vertices item for planets verticies not yet set!\n");
		return -1;
	}
	vertex >>= 2;
	*x = ri->u.vertices.x[vertex];
	*y = ri->u.vertices.y[vertex];
	*z = ri->u.vertices.z[vertex];
#ifdef DEBUG_VERTEX
	printf("using %s coordinates: vertex=%d, x=%.0f, y=%.0f, z=%.0f\n", what, vertex, *x, *y, *z);
#endif

	return 0;
}

static int use_interior_coord(const char __attribute__((unused)) *what, uint32_t vertex, int level, double *x, double *y, double *z)
{
	if ((vertex & 3)) {
		print_info("Vertex is not a multiple of four!\n");
		return -1;
	}
	if (vertex >= MAX_INTERIOR_VERTEX) {
		print_info("Vertex %d exceeds maximum vertex number %d!\n", vertex, MAX_VERTEX);
		return -1;
	}
	if (level < 1 || level > 4) {
		print_info("Level %d is out of range (1..4)!\n", level);
		return -1;
	}
	if (!render_item_vertices_interior) {
		print_info("Vertices item for interior verticies not yet set!\n");
		return -1;
	}
	vertex >>= 2;
	*x = render_item_vertices_interior->u.vertices_interior.x[vertex];
	*y = render_item_vertices_interior->u.vertices_interior.y[level - 1];
	*z = render_item_vertices_interior->u.vertices_interior.z[vertex];
	/* translate to position back to original */
	*x -= motion_new.position_east;
	*y -= motion_new.position_height;
	*z -= motion_new.position_north;
	/* translate to floating (interpolated) position offset */
	*x -= interpolation.offset_east;
	*y -= interpolation.offset_height;
	*z -= interpolation.offset_north;

#ifdef DEBUG_VERTEX
	printf("using %s coordinates: vertex=%d, x=%.0f, y=%.0f, z=%.0f\n", what, vertex, *x, *y, *z);
#endif

	return 0;
}

/* this will calculate lighting parameters for sky and shadows */
static void calculate_lighting(double sun_x, double sun_y, double sun_z, double *day, double *stars, double *shadow)
{
	double elevation;

//	azimuth = atan2(sun_z, sun_x);

	/* how high is the sun (1.0 is zenith, < 0 is below horizon) */
	elevation = atan2(sun_y, sqrt(sun_x * sun_x + sun_z * sun_z)) / (M_PI / 2);
#ifdef DEBUG_LIGHT
	printf("Elevation of SUN: %.3f\n", elevation);
#endif

	/* 'day' is how much day color is used for sky (1.0) or night color (0.0) */
	*day = (elevation - ELEVATION_SKY_NIGHT) / (ELEVATION_SKY_DAY - ELEVATION_SKY_NIGHT);
	if (*day < 0.0)
		*day = 0.0;
	if (*day > 1.0)
		*day = 1.0;
#ifdef DEBUG_LIGHT
	printf("How much day: %.3f\n", *day);
#endif

	/* 'stars' is how much we see the stars (1.0) fully bright, (0.0) no stars */
	*stars = (elevation - ELEVATION_STARS_DAY) / (ELEVATION_STARS_NIGHT - ELEVATION_STARS_DAY);
	if (*stars < 0.0)
		*stars = 0.0;
	if (*stars > 1.0)
		*stars = 1.0;
#ifdef DEBUG_LIGHT
	printf("How much stars: %.3f\n", *stars);
#endif

	/* 'shadow' is how much color (darkness) the shadow has (1.0) totally black, (0.0) not visible */
	*shadow = (elevation - ELEVATION_SHADOW_MIN) / (ELEVATION_SHADOW_MAX - ELEVATION_SHADOW_MIN);
	if (*shadow < 0.0)
		*shadow = 0.0;
	if (*shadow > 1.0)
		*shadow = 1.0;
	*shadow *= SHADOW_LEVEL;
#ifdef DEBUG_LIGHT
	printf("How much shadow: %.3f\n", *shadow);
#endif

}

/* renders one item from render list */
void render_one_item(render_item_t *render_item, int vr, enum render_shadow shadow)
{
	switch (render_item->type) {
	case RENDER_ITEM_OBJECT_INFO:
	{
#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_OBJECT_INFO object-id=%d\n", render_item->u.info.id);
#endif
		render_item_object_info = render_item;
		break;
	}
	case RENDER_ITEM_VERTICES_0:
	{
#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_VERTICES_0\n");
#endif
		render_item_vertices_0 = render_item;
		break;
	}
	case RENDER_ITEM_VERTICES_1:
	{
#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_VERTICES_1\n");
#endif
		render_item_vertices_1 = render_item;
		break;
	}
	case RENDER_ITEM_VERTICES_2:
	{
#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_VERTICES_2\n");
#endif
		render_item_vertices_2 = render_item;
		break;
	}
	case RENDER_ITEM_VERTICES_INTERIOR:
	{
#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_VERTICES_INTERIOR\n");
#endif
		if (interpolation.interior)
			render_item_vertices_interior = interpolation.interior;
		else
			render_item_vertices_interior = render_item;
		break;
	}
	case RENDER_ITEM_SKY:
	{
		double x[4], y[4], z[4];

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_SKY\n");
#endif
		/* get color */
		if (draw_shadows && ground_index >= 0) {
			uint16_t day, night;
			double day_red, day_green, day_blue;
			double night_red, night_green, night_blue;
			double day_factor;

			calculate_lighting(sun_pos_x, sun_pos_y, sun_pos_z, &day_factor, &stars_bright, &shadow_level);

			/* sky color */
			mercenary_get_sky_colors(&day, &night);
			/* shift them to the original format (atari style) */
			gamecolor2gl(&day_red, &day_green, &day_blue, day >> 1);
			gamecolor2gl(&night_red, &night_green, &night_blue, night >> 1);
			sky_color_red = day_red * day_factor + night_red * (1.0 - day_factor);
			sky_color_green = day_green * day_factor + night_green * (1.0 - day_factor);
			sky_color_blue = day_blue * day_factor + night_blue * (1.0 - day_factor);
			opengl_render_color(sky_color_red, sky_color_green, sky_color_blue, 1.0);

		} else
			opengl_render_color(render_item->u.sky.red, render_item->u.sky.green, render_item->u.sky.blue, 1.0);
		/* create box to fill view */
		x[0] = x[1] = y[1] = y[2] = -1000;
		x[2] = x[3] = y[0] = y[3] = 1000;
		z[0] = z[1] = z[2] = z[3] = 1000;
		opengl_render_polygon_and_line(x, y, z, 4);
		z[0] = z[1] = z[2] = z[3] = -1000;
		opengl_render_polygon_and_line(x, y, z, 4);
		x[0] = x[1] = z[1] = z[2] = -1000;
		x[2] = x[3] = z[0] = z[3] = 1000;
		y[0] = y[1] = y[2] = y[3] = 1000;
		opengl_render_polygon_and_line(x, y, z, 4);
		y[0] = y[1] = y[2] = y[3] = -1000;
		opengl_render_polygon_and_line(x, y, z, 4);
		y[0] = y[1] = z[1] = z[2] = -1000;
		y[2] = y[3] = z[0] = z[3] = 1000;
		x[0] = x[1] = x[2] = x[3] = 1000;
		opengl_render_polygon_and_line(x, y, z, 4);
		x[0] = x[1] = x[2] = x[3] = -1000;
		opengl_render_polygon_and_line(x, y, z, 4);
		break;
	}
	case RENDER_ITEM_GROUND:
	{
		GET_ORIENTATION;
		double x[4], y[4], z[4];
		int i;

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_GROUND\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.ground.red, render_item->u.ground.green, render_item->u.ground.blue, 1.0);
		yaw = 0.0; /* no need to rotate x-z plane, we don't want look at a corner of the 'ground-square' */
		/* create huge square */
		x[0] = x[1] = z[1] = z[2] = -10000000;
		x[2] = x[3] = z[0] = z[3] = 10000000;
		y[0] = y[1] = y[2] = y[3] = -1000;
		/* rotate vertex */
		for (i = 0; i < 4; i++)
			rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
		/* render */
		opengl_render_polygon(x, y, z, 4, 0); /* no culling, because gound is always visible! */
		break;
	}
	case RENDER_ITEM_OBJECT_POLYGON:
	case RENDER_ITEM_TAG_POLYGON_OBJECT:
	case RENDER_ITEM_TAG_POLYGON_OTHER:
	case RENDER_ITEM_ISLAND_POLYGON:
	{
		GET_ORIENTATION;
		int fix = 0;
		double x[MAX_POLYGON], y[MAX_POLYGON], z[MAX_POLYGON];
		int i, o;
		int rc;

		if (render_item->type == RENDER_ITEM_ISLAND_POLYGON) {
			if (shadow != RENDER_BEFORE_SHADOW)
				break;
		} else {
			if (shadow != RENDER_AFTER_SHADOW)
				break;
		}

#ifdef DEBUG_ITEMS
		if (render_item->type == RENDER_ITEM_OBJECT_POLYGON)
			printf("RENDER_ITEM_OBJECT_POLYGON\n");
		if (render_item->type == RENDER_ITEM_TAG_POLYGON_OBJECT)
			printf("RENDER_ITEM_TAG_POLYGON_OBJECT\n");
		if (render_item->type == RENDER_ITEM_TAG_POLYGON_OTHER)
			printf("RENDER_ITEM_TAG_POLYGON_OTHER\n");
		if (render_item->type == RENDER_ITEM_ISLAND_POLYGON)
			printf("RENDER_ITEM_ISLAND_POLYGON\n");
#endif

		/* special case where we don't want to interpolate motion (taxi/busses/intercity) */
		if ((render_item->type == RENDER_ITEM_OBJECT_POLYGON || render_item->type == RENDER_ITEM_TAG_POLYGON_OBJECT) && render_item_object_info && !render_item_object_info->u.info.moving) {
//			GET_ORIENTATION_FIX;
			fix = 1;
		}

		/* get color */
		opengl_render_color(render_item->u.polygon.red, render_item->u.polygon.green, render_item->u.polygon.blue, debug_opacity);
		/* get and rotate vertex */
		for (i = 0; i < render_item->u.polygon.vertices; i++) {
			/* get vertex */
			rc = use_coord("object", render_item->u.polygon.vertex[i], &x[i], &y[i], &z[i], fix);
			if (rc < 0)
				break;
			/* fixed objects are pre-scaled by 16, so we correct this */
			if (fix) {
				x[i] /= FIX_OBJECT_SCALE;
				y[i] /= FIX_OBJECT_SCALE;
				z[i] /= FIX_OBJECT_SCALE;
			}
			/* interpolate motion, if object is moving */
			if ((render_item->type == RENDER_ITEM_OBJECT_POLYGON || render_item->type == RENDER_ITEM_TAG_POLYGON_OBJECT)  && render_item_object_info && render_item_object_info->u.info.moving) {
				for (o = 0; o < interpolation.object_count; o++) {
					if (interpolation.object_id[o] == render_item_object_info->u.info.id)
						break;
				}
				if (o < interpolation.object_count) {
					x[i] += interpolation.object_offset_east[o];
					y[i] += interpolation.object_offset_height[o];
					z[i] += interpolation.object_offset_north[o];
				}
			}
			/* rotate vertex */
			rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
		}
		/* render */
		opengl_render_polygon(x, y, z, render_item->u.polygon.vertices, 1); /* back face culling */
		break;
	}
	case RENDER_ITEM_OBJECT_LINE:
	case RENDER_ITEM_TAG_LINE_OBJECT:
	case RENDER_ITEM_TAG_LINE_OTHER:
	{
		GET_ORIENTATION;
		int fix = 0;
		double x[2], y[2], z[2];
		int i, o;
		int rc;

		if (shadow != RENDER_AFTER_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		if (render_item->type == RENDER_ITEM_OBJECT_LINE)
			printf("RENDER_ITEM_OBJECT_LINE\n");
		if (render_item->type == RENDER_ITEM_TAG_LINE_OBJECT)
			printf("RENDER_ITEM_TAG_LINE_OBJECT\n");
		if (render_item->type == RENDER_ITEM_TAG_LINE_OTHER)
			printf("RENDER_ITEM_TAG_LINE_OTHER\n");
#endif

		/* special case where we don't want to interpolate motion (taxi/busses/intercity) */
		if ((render_item->type == RENDER_ITEM_OBJECT_LINE || render_item->type == RENDER_ITEM_TAG_LINE_OBJECT) && render_item_object_info && !render_item_object_info->u.info.moving) {
//			GET_ORIENTATION_FIX;
			fix = 1;
		}

		/* get color */
		opengl_render_color(render_item->u.line.red, render_item->u.line.green, render_item->u.line.blue, debug_opacity);
		/* get and rotate vertex */
		for (i = 0; i < 2; i++) {
			/* get vertex */
			rc = use_coord("object", render_item->u.line.vertex[i], &x[i], &y[i], &z[i], fix);
			if (rc < 0)
				break;
			/* fixed objects are pre-scaled by 16, so we correct this */
			if (fix) {
				if (render_item->type == RENDER_ITEM_TAG_LINE_OBJECT) {
					x[i] /= FIX_OBJECT_SCALE_TAG;
					y[i] /= FIX_OBJECT_SCALE_TAG;
					z[i] /= FIX_OBJECT_SCALE_TAG;
				} else {
					x[i] /= FIX_OBJECT_SCALE;
					y[i] /= FIX_OBJECT_SCALE;
					z[i] /= FIX_OBJECT_SCALE;
				}
			}
			/* interpolate motion, if object is moving */
			if ((render_item->type == RENDER_ITEM_OBJECT_LINE || render_item->type == RENDER_ITEM_TAG_LINE_OBJECT)  && render_item_object_info && render_item_object_info->u.info.moving) {
				for (o = 0; o < interpolation.object_count; o++) {
					if (interpolation.object_id[o] == render_item_object_info->u.info.id)
						break;
				}
				if (o < interpolation.object_count) {
					x[i] += interpolation.object_offset_east[o];
					y[i] += interpolation.object_offset_height[o];
					z[i] += interpolation.object_offset_north[o];
				}
			}
			/* rotate vertex */
			rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
		}
		/* render */
		opengl_render_line(x[0], y[0], z[0], x[1], y[1], z[1], 0.0);
		break;
	}
	case RENDER_ITEM_BEACON_POINT:
	{
		GET_ORIENTATION;
		double x, y, z;
		int rc;

		if (shadow != RENDER_AFTER_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_BEACON_POINT\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.point.red, render_item->u.point.green, render_item->u.point.blue, debug_opacity);
		/* get vertex */
		rc = use_coord("beacon", render_item->u.point.vertex, &x, &y, &z, 0);
		if (rc < 0)
			break;
		/* rotate vertex */
		rotate_coordinate(roll, pitch, yaw, &x, &y, &z);
		/* render */
		opengl_render_point(x, y, z, 0.0);
		break;
	}
	case RENDER_ITEM_BUILDING_EXTERIOR_POLYGON:
	{
		GET_ORIENTATION;
		double x[MAX_POLYGON], y[MAX_POLYGON], z[MAX_POLYGON];
		int i;
		int rc;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_BUILDING_EXTERIOR_POLYGON\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.polygon.red, render_item->u.polygon.green, render_item->u.polygon.blue, debug_opacity);
		/* get and rotate vertex */
		for (i = 0; i < render_item->u.polygon.vertices; i++) {
			/* get vertex */
			rc = use_coord("building exterior", render_item->u.polygon.vertex[i], &x[i], &y[i], &z[i], 0);
			if (rc < 0)
				break;
			/* rotate vertex */
			rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
		}
		/* render */
		opengl_render_polygon(x, y, z, render_item->u.polygon.vertices, 1); /* back face culling */
		break;
	}
	case RENDER_ITEM_BUILDING_EXTERIOR_LINE:
	{
		GET_ORIENTATION;
		double x[2], y[2], z[2];
		int i;
		int rc;

		if (shadow != RENDER_AFTER_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_BUILDING_EXTERIOR_LINE\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.line.red, render_item->u.line.green, render_item->u.line.blue, debug_opacity);
		/* get and rotate vertex */
		for (i = 0; i < 2; i++) {
			/* get vertex */
			rc = use_coord("building exterior", render_item->u.line.vertex[i], &x[i], &y[i], &z[i], 0);
			if (rc < 0)
				break;
			/* rotate vertex */
			rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
		}
		/* render */
		opengl_render_line(x[0], y[0], z[0], x[1], y[1], z[1], 0.0);
		break;
	}
	case RENDER_ITEM_BUILDING_INTERIOR_1TO4:
	{
		GET_ORIENTATION;
		double x[4], y[4], z[4];
		int i;
		int rc;

		if (shadow != RENDER_AFTER_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_BUILDING_INTERIOR_1TO4\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.interior14.red, render_item->u.interior14.green, render_item->u.interior14.blue, debug_opacity);
		/* get and rotate vertex */
		for (i = 0; i < 4; i++) {
			/* get vertex */
			rc = use_interior_coord("building exterior", render_item->u.interior14.vertex[i], render_item->u.interior14.level, &x[i], &y[i], &z[i]);
			if (rc < 0)
				break;
			/* rotate vertex */
			rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
		}
		/* render */
		opengl_render_polygon(x, y, z, 4, 1); /* back face culling */
		break;
	}
	case RENDER_ITEM_BUILDING_INTERIOR_5TO6:
	{
		GET_ORIENTATION;
		double x[4], y[4], z[4];
		int i;
		int vertex, level;
		int rc;

		if (shadow != RENDER_AFTER_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_BUILDING_INTERIOR_5TO6\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.interior56.red, render_item->u.interior56.green, render_item->u.interior56.blue, debug_opacity);
		/* get and rotate vertex */
		for (i = 0; i < 4; i++) {
			/* get vertex */
			vertex = (i == 0 || i == 3) ? render_item->u.interior56.vertex14 : render_item->u.interior56.vertex23;
			level = (i == 0 || i == 1) ? render_item->u.interior56.level12 : render_item->u.interior56.level34;
			rc = use_interior_coord("building interior", vertex, level, &x[i], &y[i], &z[i]);
			if (rc < 0)
				break;
			/* rotate vertex */
			rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
		}
		/* render */
		opengl_render_polygon(x, y, z, 4, 1); /* back face culling */
		break;
	}
	case RENDER_ITEM_BUILDING_INTERIOR_WALL:
	{
		GET_ORIENTATION;
		double x[4], y[4], z[4];
		int i;
		int vertex, level;
		int rc;

		if (shadow != RENDER_AFTER_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_BUILDING_INTERIOR_WALL\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.interior56.red, render_item->u.interior56.green, render_item->u.interior56.blue, debug_opacity);
		/* chedck if wall is a rectangle or a line */
		if (render_item->u.interior56.vertex14 != render_item->u.interior56.vertex23) {
			/* get and rotate vertex */
			for (i = 0; i < 4; i++) {
				/* get vertex */
				vertex = (i == 0 || i == 3) ? render_item->u.interior56.vertex14 : render_item->u.interior56.vertex23;
				level = (i == 0 || i == 1) ? render_item->u.interior56.level12 : render_item->u.interior56.level34;
				rc = use_interior_coord("building interior", vertex, level, &x[i], &y[i], &z[i]);
				if (rc < 0)
					break;
				/* rotate vertex */
				rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
			}
			/* render */
			opengl_render_polygon_and_line(x, y, z, 4); /* no culling, because walls are always visible! */
		} else {
			/* get and rotate vertex */
			for (i = 0; i < 2; i++) {
				/* get vertex */
				vertex = render_item->u.interior56.vertex14;
				level = (i == 0) ? render_item->u.interior56.level12 : render_item->u.interior56.level34;
				rc = use_interior_coord("building interior", vertex, level, &x[i], &y[i], &z[i]);
				if (rc < 0)
					break;
				/* rotate vertex */
				rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
			}
			/* render */
			opengl_render_line(x[0], y[0], z[0], x[1], y[1], z[1], 0.0);
		}
		break;
	}
	case RENDER_ITEM_COMET_POLYGON:
	{
		GET_ORIENTATION;
		double inclination = interpolation.planet_inclination, azimuth = interpolation.planet_azimuth;
		double x[MAX_POLYGON], y[MAX_POLYGON], z[MAX_POLYGON];
		int i;
		int rc;
		double rotate_sky = 0.0;

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_COMET_POLYGON\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.polygon.red, render_item->u.polygon.green, render_item->u.polygon.blue, debug_opacity);
		/* get and rotate vertex */
		for (i = 0; i < render_item->u.polygon.vertices; i++) {
			/* get vertex */
			rc = use_planet_coord("comet tail", render_item->u.polygon.vertex[i], &x[i], &y[i], &z[i]);
			if (rc < 0)
				break;
			/* rotate vertex */
			if (motion_new.planet_rotation) {
				if (fix_sky_rotation)
					rotate_sky = M_PI;
				rotate_coordinate(0.0, inclination, azimuth, &x[i], &y[i], &z[i]);
			}
			rotate_coordinate(roll, pitch, yaw + rotate_sky, &x[i], &y[i], &z[i]);
		}
		/* render */
		opengl_render_polygon_and_line(x, y, z, render_item->u.polygon.vertices); /* no culling, because we render only two (out of four) planes! */
		break;
	}
	case RENDER_ITEM_ROAD_LINE:
	{
		GET_ORIENTATION;
		double x[2], y[2], z[2];
		int i;
		int rc;

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_ROAD_LINE\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.line.red, render_item->u.line.green, render_item->u.line.blue, debug_opacity);
		/* get and rotate vertex */
		for (i = 0; i < 2; i++) {
			/* get vertex */
			rc = use_coord("road", render_item->u.line.vertex[i], &x[i], &y[i], &z[i], 0);
			if (rc < 0)
				break;
			/* rotate vertex */
			rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
		}
		/* render */
		opengl_render_line(x[0], y[0], z[0], x[1], y[1], z[1], 0.0);
		break;
	}
	case RENDER_ITEM_ROAD_POLYGON:
	{
		GET_ORIENTATION;
		double x[MAX_POLYGON], y[MAX_POLYGON], z[MAX_POLYGON];
		int i, v;
		uint32_t vertex, vertex_prev, vertex_next;
		double x_current, y_current, z_current;
		double x_prev, y_prev, z_prev;
		double x_next, y_next, z_next;
		int vertices_num;
		int rc;

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_ROAD_POLYGON\n");
#endif
		/* get color */
		opengl_render_color(render_item->u.polygon.red, render_item->u.polygon.green, render_item->u.polygon.blue, debug_opacity);
		/* get and rotate vertex */
		i = 0;
		vertices_num = render_item->u.polygon.vertices;
		for (v = 0; v < vertices_num; v++) {
			/* get vertex */
			vertex = render_item->u.polygon.vertex[v];
			rc = use_coord("road/place", vertex, &x_current, &y_current, &z_current, 0);
			if (rc < 0)
				break;
			/* check for road extension, so we extend the road to the given end point */
			if (extend_roads && vertex >= 0xf0) {
				/* previous vertex */
				vertex_prev = render_item->u.polygon.vertex[(v + vertices_num - 1) % vertices_num];
				rc = use_coord("road/place", vertex_prev, &x_prev, &y_prev, &z_prev, 0);
				if (rc < 0)
					break;
				/* next vertex */
				vertex_next = render_item->u.polygon.vertex[(v + 1) % vertices_num];
				rc = use_coord("road/place", vertex_next, &x_next, &y_next, &z_next, 0);
				if (rc < 0)
					break;
				/* extend vertices to end point position
				 * change x or z coordinate, whatever is greater
				*/
				if (fabs(x_prev - x_current) > fabs(z_prev - z_current))
					x_prev = x_next = x_current;
				else
					z_prev = z_next = z_current;
				/* store vertices */
				x[i] = x_prev;
				y[i] = y_prev;
				z[i] = z_prev;
				/* rotate vertex */
				rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
				if (i++ == MAX_POLYGON)
					break;
				x[i] = x_next;
				y[i] = y_next;
				z[i] = z_next;
				/* rotate vertex */
				rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
				if (i++ == MAX_POLYGON)
					break;
				continue;
			} else {
				/* no extension, just keep the current point as is */
				x[i] = x_current;
				y[i] = y_current;
				z[i] = z_current;
				/* rotate vertex */
				rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
				if (i++ == MAX_POLYGON)
					break;
			}
		}
		/* render */
		opengl_render_polygon(x, y, z, i, 0); /* no culling, because polygons lay on the gound and are always visible! */
		break;
	}
	case RENDER_ITEM_PLANET:
	{
		GET_ORIENTATION;
		double front_red, front_green, front_blue, back_red, back_green, back_blue;
		double inclination = interpolation.planet_inclination, azimuth = interpolation.planet_azimuth;
		double sun_x, sun_y, sun_z, angle_sun_h;
		double loc_x, loc_y, loc_z, angle_loc_h, angle_loc_v;
		double circle_x[PLANET_VERTICES], circle_y[PLANET_VERTICES], circle_z[PLANET_VERTICES];
		double crescent_x[PLANET_VERTICES], crescent_y[PLANET_VERTICES], crescent_z[PLANET_VERTICES];
		double x[PLANET_VERTICES], y[PLANET_VERTICES], z[PLANET_VERTICES];
		double dist, size, angle, fabs_angle, crescent;
		double _sin, _cos;
		int i;
		int rc;
		double rotate_sky = 0.0;

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_PLANET\n");
#endif

		/* color of planets */
		front_red = render_item->u.planet.front_red;
		front_green = render_item->u.planet.front_green;
		front_blue = render_item->u.planet.front_blue;
		back_red = render_item->u.planet.back_red;
		back_green = render_item->u.planet.back_green;
		back_blue = render_item->u.planet.back_blue;
		if (draw_shadows && ground_index >= 0) {
			/* front color is blended by sky */
			front_red = front_red * (1.0 - sky_color_red) + sky_color_red;
			front_green = front_green * (1.0 - sky_color_green) + sky_color_green;
			front_blue = front_blue * (1.0 - sky_color_blue) + sky_color_blue;
			back_red = sky_color_red * (1.0 - stars_bright);
			back_green = sky_color_green * (1.0 - stars_bright);
			back_blue = sky_color_blue * (1.0 - stars_bright);
		}

		/* get location */
		rc = use_planet_coord("planet(sun)", 0, &sun_x, &sun_y, &sun_z);
		if (rc < 0)
			break;
		rc = use_planet_coord("planet", render_item->u.planet.vertex, &loc_x, &loc_y, &loc_z);
		if (rc < 0)
			break;
		/* get size */
		size = render_item->u.planet.size;
		/* rotate vertex */
		if (motion_new.planet_rotation) {
			if (fix_sky_rotation)
				rotate_sky = M_PI;
			rotate_coordinate(0.0, inclination, azimuth, &sun_x, &sun_y, &sun_z);
			rotate_coordinate(0.0, inclination, azimuth, &loc_x, &loc_y, &loc_z);
		}
		/* store sun location */
		sun_valid = 1;
		sun_pos_x = sun_x;
		sun_pos_y = sun_y;
		sun_pos_z = sun_z;
		/* rotate sun */
		rotate_coordinate(roll, pitch, yaw + rotate_sky, &sun_x, &sun_y, &sun_z);
		rotate_coordinate(roll, pitch, yaw + rotate_sky, &loc_x, &loc_y, &loc_z);

		/* distance to planet */
		dist = sqrt(loc_x * loc_x + loc_y * loc_y + loc_z * loc_z);

		/* calculate direction of the sun */
		angle_sun_h = atan2(sun_x, sun_z);
		angle_loc_h = atan2(loc_x, loc_z);
		angle_loc_v = atan2(loc_y, sqrt(loc_x * loc_x + loc_z * loc_z));
		/* angle between planets */
		angle = angle_sun_h - angle_loc_h;
		if (angle > M_PI)
			angle -= 2.0 * M_PI;
		if (angle < -M_PI)
			angle += 2.0 * M_PI;
		/* absolute angle to be used as crescent */
		fabs_angle = fabs(angle);
		if (fabs_angle > M_PI / 2.0)
			fabs_angle = M_PI - fabs_angle;

		/* on which side are we (sun is always bright, vertex == 0) */
		if ((angle < M_PI / 2.0 && angle > -M_PI / 2.0) || render_item->u.planet.vertex == 0) {
			/* get front side color */
			opengl_render_color(front_red, front_green, front_blue, debug_opacity);
		} else {
			/* get back side color */
			opengl_render_color(back_red, back_green, back_blue, debug_opacity);
		}

		/* create and render cicle */
		crescent = sin((1.0 - fabs_angle / (M_PI / 2)) * (M_PI / 2.0));
		if (vr) {
			for (i = 0; i < PLANET_VERTICES; i++) {
				_sin = size * sin(2.0 * M_PI * (double)i / PLANET_VERTICES);
				_cos = size * cos(2.0 * M_PI * (double)i / PLANET_VERTICES);
				circle_x[i] = _sin * planet_aspect;
				circle_y[i] = _cos;
				circle_z[i] = 0.0;
				crescent_x[i] = circle_x[i] * crescent;
				crescent_y[i] = circle_y[i];
				crescent_z[i] = circle_z[i];
				/* rotate circle and cresent towards observer (billboarding) */
				rotate_coordinate(0.0, -angle_loc_v, 0.0, &circle_x[i], &circle_y[i], &circle_z[i]);
				rotate_coordinate(0.0, 0.0, angle_loc_h, &circle_x[i], &circle_y[i], &circle_z[i]);
				rotate_coordinate(0.0, -angle_loc_v, 0.0, &crescent_x[i], &crescent_y[i], &crescent_z[i]);
				rotate_coordinate(0.0, 0.0, angle_loc_h, &crescent_x[i], &crescent_y[i], &crescent_z[i]);
			}
		} else {
			for (i = 0; i < PLANET_VERTICES; i++) {
				_sin = size * sin(2.0 * M_PI * (double)i / PLANET_VERTICES);
				_cos = size * cos(2.0 * M_PI * (double)i / PLANET_VERTICES);
				circle_x[i] = _sin * planet_aspect;
				circle_y[i] = _cos;
				circle_z[i] = 0.0;
				crescent_x[i] = circle_x[i] * crescent;
				crescent_y[i] = circle_y[i];
				crescent_z[i] = circle_z[i];
			}
		}
		for (i = 0; i < PLANET_VERTICES; i++) {
			x[i] = (loc_x + circle_x[i]) / dist * 1000000.0;
			y[i] = (loc_y + circle_y[i]) / dist * 1000000.0;
			z[i] = (loc_z + circle_z[i]) / dist * 1000000.0;
		}
		opengl_render_polygon_and_line(x, y, z, PLANET_VERTICES); /* no culling, its a planet! */

		if (render_item->u.planet.vertex == 0) {
			/* sun has no crescent */
			break;
		}

		/* on which side are we */
		if (angle < M_PI / 2.0 && angle > -M_PI / 2.0) {
			/* get back side color */
			opengl_render_color(back_red, back_green, back_blue, debug_opacity);
		} else {
			/* get front side color */
			opengl_render_color(front_red, front_green, front_blue, debug_opacity);
		}

		/* create and render crescent */
		if (angle > M_PI / 2.0 || (angle < 0.0 && angle > -M_PI / 2.0)) {
			/* to the right */
			for (i = 0; i < PLANET_VERTICES / 2 + 1; i++) {
				x[i] = (loc_x + circle_x[i]) / dist * 1000000.0;
				y[i] = (loc_y + circle_y[i]) / dist * 1000000.0;
				z[i] = (loc_z + circle_z[i]) / dist * 1000000.0;
			}
			_sin = sin((1.0 - angle / (M_PI / 2)) * (M_PI / 2.0));
			for (; i < PLANET_VERTICES; i++) {
				x[i] = (loc_x + crescent_x[i]) / dist * 1000000.0;
				y[i] = (loc_y + crescent_y[i]) / dist * 1000000.0;
				z[i] = (loc_z + crescent_z[i]) / dist * 1000000.0;
			}
		} else {
			/* to the left */
			for (i = 0; i < PLANET_VERTICES / 2 + 1; i++) {
				x[i] = (loc_x + crescent_x[i]) / dist * 1000000.0;
				y[i] = (loc_y + crescent_y[i]) / dist * 1000000.0;
				z[i] = (loc_z + crescent_z[i]) / dist * 1000000.0;
			}
			for (; i < PLANET_VERTICES; i++) {
				x[i] = (loc_x + circle_x[i]) / dist * 1000000.0;
				y[i] = (loc_y + circle_y[i]) / dist * 1000000.0;
				z[i] = (loc_z + circle_z[i]) / dist * 1000000.0;
			}
		}
		opengl_render_polygon_and_line(x, y, z, PLANET_VERTICES); /* no culling, its a planet! */
		opengl_render_point(loc_x, loc_y, loc_z, 0.0); /* planet is visible at any distance - at least as a point  */
		break;
	}
	case RENDER_ITEM_STARS:
	{
		double tilt_offset = 0;
		double x_offset = 0;
		uint16_t color[16];
		double view_width, yaw = interpolation.orientation_raw_yaw;
		double pitch = interpolation.orientation_raw_pitch;
		uint32_t table, table_start;
		double x, y, z;
		int i;
		double red, green, blue;
		double rotate_sky = 0.0;

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_STARS\n");
#endif

		/* no stars at day */
		if (stars_bright == 0.0)
			break;

		/* use default fov of 64 to calculate z distance */
		z = 160.0 / frustum_slope_64;

		view_width = (int)(320.0 / frustum_slope_64 * frustum_slope_fov);

		/* get palette */
		for (i = 0; i < 16; i++)
			color[i] = m68k_read_memory_16(mercenary_palette_stars() + (i << 2));

		if (!vr && !improve_stars) {
			/* render legacy stars (as with the original game) */

			/* table offset is 91, so we substract it and add it back with different FOV, so table begins at later
			 * full turn is 1024, we have default of 64 degrees: 1024 / 360 * 64 = 182
			 * then we half it, so we get to the center via 91
			 */

			if (render_item->u.stars.above_zenith)
				yaw = 0x200 + yaw;
			yaw = fmod(yaw + 91.0 - (91.0 * (frustum_slope_fov / frustum_slope_64)) + 65536.0, 0x400);
			yaw *= 2.0;
			table = mercenary_star_table();
			table_start = table + m68k_read_memory_16(table);
			table += m68k_read_memory_16(table + ((uint32_t)yaw & 0x7fe));
			yaw = yaw / (double)0x800 * 1800.0;

			if (render_item->u.stars.above_zenith)
				pitch = 0x200 - pitch;
			pitch = fmod(pitch + 65536.0, 0x400);
			pitch -= render_item->u.stars.v_offset;
			pitch *= 4;
			pitch = pitch * (double)0x6ccc / 65536.0;

			while (1) {
				x = m68k_read_memory_16(table);
				if (x >= 1800.0) {
					x_offset += 1800.0;
					table = table_start;
				}
				x = (view_width - 1) - m68k_read_memory_16(table) - x_offset + yaw;
				table += 2;
				if (x < 0.0)
					break;
				/* special case where we tilt the view when flying on the planet */
				if (render_item->u.stars.tilt) {
					/* use offset as given by game: 160 is half of the screen width
					 * we extend the width to the actual FOV, so it fits
					 */
					tilt_offset = ((x - (160.0 / frustum_slope_64 * frustum_slope_fov)) * render_item->u.stars.tilt_value) / 65536.0;
				}
				y = (double)((m68k_read_memory_16(table)) & 0x1ff) - pitch + tilt_offset;
				table += 2;
				if (render_item->u.stars.above_zenith) {
					x = (double)(view_width - 1) - x;
					y = -1 - y + 136;
				}
				/* get color */
				gamecolor2gl(&red, &green, &blue, color[(m68k_read_memory_8(table - 2) & 0x3c) >> 2]);
				opengl_render_color(red, green, blue, debug_opacity * stars_bright);
				/* render point */
				opengl_render_point(160.0 / frustum_slope_64 * frustum_slope_fov - x, 68.0 - y, z, 0.0);
			}
		} else {
			/* render ovr stars, render star table as a sphere */

			double h, v1, v2;
			GET_ORIENTATION;
			double inclination = interpolation.planet_inclination, azimuth = interpolation.planet_azimuth;

			if (render_item->u.stars.above_zenith)
				break;

			table = mercenary_star_table();
			table += m68k_read_memory_16(table);

			while (1) {
				x = m68k_read_memory_16(table);
				if (x >= 1800.0)
					break;
				table += 2;
				y = (double)((m68k_read_memory_16(table)) & 0x1ff) - 108.796875;
				table += 2;
				/* get color */
				gamecolor2gl(&red, &green, &blue, color[(m68k_read_memory_8(table - 2) & 0x3c) >> 2]);
				opengl_render_color(red, green, blue, debug_opacity * stars_bright);
				/* render point */
				h = (900.0 - x + 160) / 900.0 * M_PI;
				v1 = (68.0 - y) / 900.0 * M_PI;
				/* wrap star field (is actually 86.2 degrees high) */
				if (v1 <= 0.0)
					v2 = v1 + (86.2 / 180.0 * M_PI);
				else
					v2 = v1 - (86.2 / 180.0 * M_PI);
				if (v1 < 0.934 && v1 > -0.934) {
					/* be sure that v1 will not exceed PI/2 */
					v1 = v1 / cos(v1); /* FIXME: there should be a better way to distribute stars equally */
					x = -sin(h) * cos(v1);
					y = sin(v1);
					z = cos(h) * cos(v1);
					if (motion_new.planet_rotation) {
						if (fix_sky_rotation)
							rotate_sky = M_PI;
						rotate_coordinate(0.0, inclination, azimuth, &x, &y, &z);
					}
					rotate_coordinate(roll, pitch, yaw + rotate_sky, &x, &y, &z);
					opengl_render_point(1000000.0 * x, 1000000.0 * y, 1000000.0 * z, 0.0);
				}
				if (v2 < 0.934 && v2 > -0.934) {
					/* be sure that v2 will not exceed PI/2 */
					v2 = v2 / cos(v2) /* FIXME: there should be a better way to distribute stars equally */;
					x = -sin(h) * cos(v2);
					y = sin(v2);
					z = cos(h) * cos(v2);
					if (motion_new.planet_rotation) {
						if (fix_sky_rotation)
							rotate_sky = M_PI;
						rotate_coordinate(0.0, inclination, azimuth, &x, &y, &z);
					}
					rotate_coordinate(roll, pitch, yaw + rotate_sky, &x, &y, &z);
					opengl_render_point(1000000.0 * x, 1000000.0 * y, 1000000.0 * z, 0.0);
				}
			}
		}
		break;
	}
	case RENDER_ITEM_INTERSTELLAR_STARS:
	{
		uint16_t color[16];
		double z;
		int i;
		double red, green, blue;

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_INTERSTELLAR_STARS\n");
#endif
		/* use default fov of 64 to calculate z distance */
		z = 160.0 / frustum_slope_64;

		/* get palette */
		for (i = 0; i < 16; i++)
			color[i] = m68k_read_memory_16(mercenary_palette_stars() + (i << 2));

		for (i = 0; i < render_item->u.interstars.count; i++) {
			gamecolor2gl(&red, &green, &blue, color[render_item->u.interstars.color[i]]);
			opengl_render_color(red, green, blue, debug_opacity);
			/* render point */
			opengl_render_point((160.0 - (double)render_item->u.interstars.x[i]) * 100, (68.0 - (double)render_item->u.interstars.y[i]) * 100, z * 100, 0.0);
		}
		break;
	}
	case RENDER_ITEM_INTERSTELLAR_SUN:
	{
		double red, green, blue;

		if (shadow != RENDER_BEFORE_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_INTERSTELLAR_SUN\n");
#endif
		/* white */
		gamecolor2gl(&red, &green, &blue, 0x777);
		opengl_render_color(red, green, blue, debug_opacity);
		/* render point */
		opengl_render_point(0.0, 0.0, 1000000.0, 0.0);
		break;
	}
	case RENDER_ITEM_SIGHTS:
	{
		double x[4], y[4], z[4];
		double red, green, blue;

		if (shadow != RENDER_AFTER_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_SIGHTS\n");
#endif
		/* use default fov of 64 to calculate z distance */
		z[0] = z[1] = z[2] = z[3] = SIGHT_DIST;

		/* white */
		gamecolor2gl(&red, &green, &blue, 0x777);
		opengl_render_color(red, green, blue, debug_opacity);

		y[0] = y[3] = -1.0 / 160.0 * SIGHT_DIST * frustum_slope_64;
		y[1] = y[2] = 1.0 / 160.0 * SIGHT_DIST * frustum_slope_64;
		x[0] = x[1] = -16.0 / 160.0 * SIGHT_DIST * frustum_slope_64;
		x[2] = x[3] = -8.0 / 160.0 * SIGHT_DIST * frustum_slope_64;
		opengl_render_polygon(x, y, z, 4, 0); /* no culling, because sights are always visible! */
		x[0] = x[1] = 8.0 / 160.0 * SIGHT_DIST * frustum_slope_64;
		x[2] = x[3] = 16.0 / 160.0 * SIGHT_DIST * frustum_slope_64;
		opengl_render_polygon(x, y, z, 4, 0); /* no culling, because sights are always visible! */
		break;
	}
	case RENDER_ITEM_EXPLOSION:
	{
		GET_ORIENTATION;
		double loc_x, loc_y, loc_z, size;
		double x[EXPLOSION_VERTICES], y[EXPLOSION_VERTICES], z[EXPLOSION_VERTICES];
		int i, e;

		if (shadow != RENDER_AFTER_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		printf("RENDER_ITEM_EXPLOSION\n");
#endif
		opengl_render_color(render_item->u.explosion.red, render_item->u.explosion.green, render_item->u.explosion.blue, debug_opacity);

		for (e = 0; e < render_item->u.explosion.count; e++) {
			loc_x = render_item->u.explosion.x[e];
			loc_y = render_item->u.explosion.y[e];
			loc_z = render_item->u.explosion.z[e];
			/* rotate vertex */
			rotate_coordinate(roll, pitch, yaw, &loc_x, &loc_y, &loc_z);
			/* calculate size from projected size and z (256 is the distance of the game's projection plane) */
			size = render_item->u.explosion.size[e] * (fabs(loc_z) / 256.0);
			/* create and render cicle */
			for (i = 0; i < EXPLOSION_VERTICES; i++) {
				x[i] = loc_x + size * sin(2 * M_PI * (double)i / EXPLOSION_VERTICES) * explosion_aspect;
				y[i] = loc_y + size * cos(2 * M_PI * (double)i / EXPLOSION_VERTICES);
				z[i] = loc_z;
			}
			opengl_render_polygon_and_line(x, y, z, EXPLOSION_VERTICES); /* no culling, its a debris! */
			opengl_render_point(loc_x, loc_y, loc_z, 0.0); /* debris is visible at any distance - at least as a point  */
		}
		break;
	}
	case RENDER_ITEM_SHADOW_BUILDING_POLYGON:
	case RENDER_ITEM_SHADOW_BUILDING_LINE:
	case RENDER_ITEM_SHADOW_OBJECT_POLYGON:
	case RENDER_ITEM_SHADOW_OBJECT_LINE:
	{
		GET_ORIENTATION;
		double x[MAX_POLYGON], y[MAX_POLYGON], z[MAX_POLYGON];
		double shadow_x, shadow_z;
		int o;
		int i;

		if (shadow != RENDER_AT_SHADOW)
			break;

#ifdef DEBUG_ITEMS
		if (render_item->type == RENDER_ITEM_SHADOW_POLYGON)
			printf("RENDER_ITEM_SHADOW_POLYGON\n");
		if (render_item->type == RENDER_ITEM_SHADOW_LINE)
			printf("RENDER_ITEM_SHADOW_LINE\n");
#endif

		if (!sun_valid || sun_pos_y < 1.0)
			return;

		if (shadow_level == 0.0)
			return;

		/* calculate shadow */
		shadow_x = -sun_pos_x / sun_pos_y;
		shadow_z = -sun_pos_z / sun_pos_y;

		opengl_render_color_alpha(0, 0, 0, shadow_level);

		if (render_item->type == RENDER_ITEM_SHADOW_BUILDING_POLYGON
		 || render_item->type == RENDER_ITEM_SHADOW_OBJECT_POLYGON) {
			/* get and rotate vertex */
			for (i = 0; i < render_item->u.shadow_polygon.vertices; i++) {
				x[i] = render_item->u.shadow_polygon.x[i];
				y[i] = render_item->u.shadow_polygon.y[i];
				z[i] = render_item->u.shadow_polygon.z[i];
				/* interpolate motion, if object is moving */
				if (render_item->type == RENDER_ITEM_SHADOW_OBJECT_POLYGON && render_item_object_info && render_item_object_info->u.info.moving) {
					for (o = 0; o < interpolation.object_count; o++) {
						if (interpolation.object_id[o] == render_item_object_info->u.info.id)
							break;
					}
					if (o < interpolation.object_count) {
						x[i] += interpolation.object_offset_east[o];
						y[i] += interpolation.object_offset_height[o];
						z[i] += interpolation.object_offset_north[o];
					}
				}
				/* flatten polygon into shadow */
				x[i] += y[i] * shadow_x;
				z[i] += y[i] * shadow_z;
				y[i] = 0.0;
				/* apply interpolation */
				x[i] -= interpolation.offset_east;
				y[i] -= interpolation.offset_height;
				z[i] -= interpolation.offset_north;
				/* apply position */
				x[i] += render_item->u.shadow_polygon.x_pos;
				y[i] += render_item->u.shadow_polygon.y_pos;
				z[i] += render_item->u.shadow_polygon.z_pos;
				/* rotate vertex */
				rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
			}
			/* render */
			opengl_render_polygon(x, y, z, render_item->u.shadow_polygon.vertices, 0); /* no back face culling */
		}
		if (render_item->type == RENDER_ITEM_SHADOW_BUILDING_LINE
		 || render_item->type == RENDER_ITEM_SHADOW_OBJECT_LINE) {
			/* get and rotate vertex */
			for (i = 0; i < 2; i++) {
				x[i] = render_item->u.shadow_line.x[i];
				y[i] = render_item->u.shadow_line.y[i];
				z[i] = render_item->u.shadow_line.z[i];
				/* interpolate motion, if object is moving */
				if (render_item->type == RENDER_ITEM_SHADOW_OBJECT_LINE && render_item_object_info && render_item_object_info->u.info.moving) {
					for (o = 0; o < interpolation.object_count; o++) {
						if (interpolation.object_id[o] == render_item_object_info->u.info.id)
							break;
					}
					if (o < interpolation.object_count) {
						x[i] += interpolation.object_offset_east[o];
						y[i] += interpolation.object_offset_height[o];
						z[i] += interpolation.object_offset_north[o];
					}
				}
				/* flatten polygon into shadow */
				x[i] += y[i] * shadow_x;
				z[i] += y[i] * shadow_z;
				y[i] = 0.0;
				/* apply interpolation */
				x[i] -= interpolation.offset_east;
				y[i] -= interpolation.offset_height;
				z[i] -= interpolation.offset_north;
				/* apply position */
				x[i] += render_item->u.shadow_line.x_pos;
				y[i] += render_item->u.shadow_line.y_pos;
				z[i] += render_item->u.shadow_line.z_pos;
				/* rotate vertex */
				rotate_coordinate(roll, pitch, yaw, &x[i], &y[i], &z[i]);
			}
			/* render */
			opengl_render_line(x[0], y[0], z[0], x[1], y[1], z[1], 0.0);
		}
		break;
	}
	default:
		print_info("Unknown render item type, please fix!\n");
	}
}

/*
 * interpolation
 */

static double interpolate_orientation(double old, double new, double inter)
{
	double turn = new - old;

	if (turn > M_PI)
		turn -= 2.0 * M_PI;
	if (turn < -M_PI)
		turn += 2.0 * M_PI;

	/* don't interpolate, if our rotation was too fast.
	 * e.g: taxi drive around corder, load/quit game, ...
	 */ 
	if (turn > M_PI / 8.0 || turn < -M_PI / 8.0)
		return new;
	
	new = old + turn * inter;
	
	if (new > M_PI)
		new -= 2.0 * M_PI;
	if (new < -M_PI)
		new += 2.0 * M_PI;

	return new;
}

static double interpolate_raw_orientation(uint16_t old, uint16_t new, double inter)
{
	int16_t turn = (new - old) & 0x3ff;

	if (turn > 0x200)
		turn -= 0x400;

	/* don't interpolate, if our rotation was too fast.
	 * e.g: taxi drive around corder, load/quit game, ...
	 */ 
	if (turn > 0x200 / 8 || turn < -0x200 / 8)
		return new;

	/* don't do modulo 0x400, since the user of this data does it */	
	return (double)old + (double)turn * inter;
}

static double interpolate_offset(int32_t old, int32_t new, double inter, int32_t limit)
{
	double offset;

	/* be sure to look only at the lower 28 bits, because on planet these bits are used only */
	if (ground_index >= 0)
		offset = wrap_int28(old - new);
	else
		offset = (int32_t)(old - new);

	if (limit > 0 && (offset > limit || offset < -limit))
		return new;

	return offset * (1.0 - inter);
}

static render_item_t *interpolate_door(double inter)
{
	static render_item_t interpolated;
	render_item_t *old_vertices = render_list_old, *new_vertices = render_list_new;
	int i, ii;

	/* find old and new vertices */
	while (old_vertices && old_vertices->type != RENDER_ITEM_VERTICES_INTERIOR)
		old_vertices = old_vertices->next;
	while (new_vertices && new_vertices->type != RENDER_ITEM_VERTICES_INTERIOR)
		new_vertices = new_vertices->next;

	/* building does not exist in old or new render */
	if (!old_vertices || !new_vertices)
		return NULL;

	memcpy(&interpolated, new_vertices, sizeof(interpolated));
	ii = MAX_INTERIOR_VERTEX >> 2;
	for (i = 0; i < ii; i++) {
		/* vertex must exist in both frames */
		if (!old_vertices->u.vertices_interior.set[i] || !new_vertices->u.vertices_interior.set[i])
			continue;
		/* all verices must not have be too far away */
		if (fabs(old_vertices->u.vertices_interior.x[i] - new_vertices->u.vertices_interior.x[i]) > 100.0)
			return NULL;
		if (fabs(old_vertices->u.vertices_interior.z[i] - new_vertices->u.vertices_interior.z[i]) > 100.0)
			return NULL;
		/* interpolate */
		interpolated.u.vertices_interior.x[i] =
		(double)old_vertices->u.vertices_interior.x[i] * (1.0 - inter) +
		(double)new_vertices->u.vertices_interior.x[i] * inter;
		interpolated.u.vertices_interior.z[i] =
		(double)old_vertices->u.vertices_interior.z[i] * (1.0 - inter) +
		(double)new_vertices->u.vertices_interior.z[i] * inter;
	}
	for (i = 0; i < 4; i++) {
		if (old_vertices->u.vertices_interior.y[i] != new_vertices->u.vertices_interior.y[i])
			return NULL;
	}

	return &interpolated;
}

/* make a list of objects that moved and store their displacement */
static void interpolate_objects(double inter)
{
	render_item_t *old_info, *new_info;
	int count;

	/* hunt for objects that exist in both (old and new) lists and moved */
	count = 0;
	for (new_info = render_list_new; new_info; new_info = new_info -> next) {
		/* not an object */
		if (new_info->type != RENDER_ITEM_OBJECT_INFO)
			continue;
		/* not moving */
		if (!new_info->u.info.moving)
			continue;
		/* interiors don't move */
		if (new_info->u.info.id < 0)
			continue;
		/* check matching object with same ID */
		for (old_info = render_list_old; old_info; old_info = old_info -> next) {
			/* not an object */
			if (old_info->type != RENDER_ITEM_OBJECT_INFO)
				continue;
			/* not moving */
			if (!old_info->u.info.moving)
				continue;
			/* same id */
			if (old_info->u.info.id == new_info->u.info.id)
				break;
		}
		/* no matching object found */
		if (!old_info)
			continue;
		/* not moving */
		if (old_info->u.info.east == new_info->u.info.east
		 && old_info->u.info.height == new_info->u.info.height
		 && old_info->u.info.north == new_info->u.info.north)
			continue;
		/* interpolate and store */
		interpolation.object_id[count] = new_info->u.info.id;
		interpolation.object_offset_east[count] = interpolate_offset(old_info->u.info.east, new_info->u.info.east, inter, 4096);
		interpolation.object_offset_height[count] = interpolate_offset(old_info->u.info.height, new_info->u.info.height, inter, 4096);
		interpolation.object_offset_north[count] = interpolate_offset(old_info->u.info.north, new_info->u.info.north, inter, 4096);
		if (count++ == MAX_MOVING_OBJECTS)
			break;
	}
	interpolation.object_count = count;
}

/* make a vertex list of interpolated planets */
static render_item_t *interpolate_planets(double inter)
{
	static render_item_t interpolated;
	render_item_t *old_info, *new_info;
	render_item_t *old_vertices = NULL, *new_vertices = NULL;
	int i;

	/* get vertices for planets/comet */
	for (old_info = render_list_old; old_info; old_info = old_info -> next) {
		if (old_info->type == RENDER_ITEM_VERTICES_0)
			old_vertices = old_info;
		/* check until comet's polygons are renderd. at this time the coordinates are all ready, so we can interpolate */
		if (old_info->type == RENDER_ITEM_COMET_POLYGON)
			break;
	}
	for (new_info = render_list_new; new_info; new_info = new_info -> next) {
		if (new_info->type == RENDER_ITEM_VERTICES_0)
			new_vertices = new_info;
		/* check until comet's polygons are renderd. at this time the coordinates are all ready, so we can interpolate */
		if (new_info->type == RENDER_ITEM_COMET_POLYGON)
			break;
	}

	/* check for existing planet's vertices and if planets are rendered in both lists (old and new) */
	if (old_info == NULL || new_info == NULL || old_vertices == NULL || new_vertices == NULL)
		return NULL;

	/* interpolate vertices */
	for (i = 0; i < (MAX_VERTEX >> 2); i++) {
		interpolated.u.vertices.x[i] = old_vertices->u.vertices.x[i] * (1.0 - inter) + new_vertices->u.vertices.x[i] * inter;
		interpolated.u.vertices.y[i] = old_vertices->u.vertices.y[i] * (1.0 - inter) + new_vertices->u.vertices.y[i] * inter;
		interpolated.u.vertices.z[i] = old_vertices->u.vertices.z[i] * (1.0 - inter) + new_vertices->u.vertices.z[i] * inter;
	}

	return &interpolated;
}

/* always renders NEW! items
 * use inter == 1.0 to render motion to vertices of NEW items
 * use inter 0.0 .. 1.0 to interpolate motion between OLD and NEW items
 * return 0, if the scene was rendered, returns < 0, if there is no scene
 */
int render_all_items(double inter, int vr)
{
	render_item_object_info = NULL;
	render_item_vertices_0 = render_item_vertices_1 = render_item_vertices_2 = NULL;
	render_item_vertices_interior = NULL;
	render_item_vertices_planets = NULL;

	/* no interpolation when leaving or entering planet to/from space */
	if ((last_ground_index < 0 && ground_index >= 0)
	 || (last_ground_index >= 0 && ground_index < 0)) {
		inter = 1.0;
	}

	/* reset interpolation */
	memset(&interpolation, 0, sizeof(interpolation));
	interpolation.orientation_roll = motion_new.orientation_roll;
	interpolation.orientation_pitch = motion_new.orientation_pitch;
	interpolation.orientation_yaw = motion_new.orientation_yaw;
	interpolation.orientation_raw_pitch = motion_new.orientation_raw_pitch;
	interpolation.orientation_raw_yaw = motion_new.orientation_raw_yaw;
	interpolation.planet_inclination = motion_new.planet_inclination;
	interpolation.planet_azimuth = motion_new.planet_azimuth;
//printf("we are at %08x %08x %08x\n", motion_new.position_east, motion_new.position_height, motion_new.position_north);

	/* do interpolation */
	if (inter != 1.0 && render_list_old) {
		/* interpolate orientation */
		interpolation.orientation_roll = interpolate_orientation(motion_old.orientation_roll, motion_new.orientation_roll, inter);
		interpolation.orientation_pitch = interpolate_orientation(motion_old.orientation_pitch, motion_new.orientation_pitch, inter);
		interpolation.orientation_yaw = interpolate_orientation(motion_old.orientation_yaw, motion_new.orientation_yaw, inter);
		interpolation.orientation_raw_pitch = interpolate_raw_orientation(motion_old.orientation_raw_pitch, motion_new.orientation_raw_pitch, inter);
		interpolation.orientation_raw_yaw = interpolate_raw_orientation(motion_old.orientation_raw_yaw, motion_new.orientation_raw_yaw, inter);
		interpolation.planet_inclination = interpolate_orientation(motion_old.planet_inclination, motion_new.planet_inclination, inter);
		interpolation.planet_azimuth = interpolate_orientation(motion_old.planet_azimuth, motion_new.planet_azimuth, inter);

		/* interpolate position */
		interpolation.offset_east = interpolate_offset(motion_old.position_east, motion_new.position_east, inter, 0);
		interpolation.offset_height = interpolate_offset(motion_old.position_height, motion_new.position_height, inter, 0);
		interpolation.offset_north = interpolate_offset(motion_old.position_north, motion_new.position_north, inter, 0);
		/* prevent glitch when using elevators: a sudden vertical move is ignored
		 * this is not the best solution, because fast vertical flying (from 0) will also be detected */
		if (old_height_offset == 0
		 && (new_height_offset >= 150 || new_height_offset <= -150)) {
			interpolation.offset_east = 0.0;
			interpolation.offset_height = 0.0;
			interpolation.offset_north = 0.0;
		}

		/* interpolate doors of building (if any) */
		interpolation.interior = interpolate_door(inter);

		/* interpolate objects */
		interpolate_objects(inter);

		/* interpolate planets */
		interpolation.planets = interpolate_planets(inter);
	}

	/* return failure, if nothing can be rendered */
	if (!render_list_new)
		return -1;

	sun_valid = 0;
	stars_bright = 1.0;

	for (render_item = render_list_new; render_item; render_item = render_item->next) {
		render_one_item(render_item, vr, RENDER_BEFORE_SHADOW);
	}
	if (draw_shadows && ground_index >= 0) {
		begin_shadow_render();
		for (render_item = render_list_new; render_item; render_item = render_item->next) {
			render_one_item(render_item, vr, RENDER_AT_SHADOW);
		}
		end_shadow_render();
	}
	for (render_item = render_list_new; render_item; render_item = render_item->next) {
		render_one_item(render_item, vr, RENDER_AFTER_SHADOW);
	}

	return 0;
}

int render_all_white(int vr)
{
	render_item_t sky_item;

	memset(&sky_item, 0, sizeof(sky_item));
	sky_item.type = RENDER_ITEM_SKY;

	/* set color */
	sky_item.u.ground.red = 1.0;
	sky_item.u.ground.green = 1.0;
	sky_item.u.ground.blue = 1.0;

	render_one_item(&sky_item, vr, RENDER_BEFORE_SHADOW);

	return 0;
}

void render_capture_reset(void)
{
	/* flush old list, if exists */
	flush_old_items();
	/* flush new list, if exists */
	render_list_old = render_list_new;
	flush_old_items();
	/* reset list pointers */
	render_list_old = NULL;
	render_list_new = NULL;
	render_list_end = &render_list_new;
	render_item = NULL;
}

int render_capture_is_interstellar(void)
{
	if (!render_list_new)
		return 0;
	for (render_item = render_list_new; render_item; render_item = render_item->next) {
		if (render_item->type == RENDER_ITEM_INTERSTELLAR_STARS)
			return 1;
	}

	return 0;
}

